################################################################################
# Licensed Materials - Property of IBM Copyright IBM Corporation 2012. All Rights Reserved.
# U.S. Government Users Restricted Rights - Use, duplication or disclosure restricted by GSA ADP
# Schedule Contract with IBM Corp.
################################################################################

require "logger" # Ruby Logging
require "devops_deployment_lib/util"
require "devops_deployment_lib/devops_server_interface"


################################################################################
# Class Definition                                                             #
################################################################################

module DevopsDeploymentLib
  
  # Objects instantiated from the Environment Class represent the
  # bindings an application will have to a certain environment. For example,
  # an application deployed to a test environment might connect to a different
  # database as opposed to when the app gets deployed to a production environment.
  #
  # The Environment class provides a simple way to create an Environment Bindings
  # hash from a JSON formatted string or file, or from an entirely other hash object.
  #
  # Once instatiated, the hash can be accessed and modified through the "bindings" class
  # attribute.
  #
  class Environment
    include DevopsDeploymentLib::Util
    
    attr_accessor :bindings
 
    
    # Create a new set of environment bindings
    #
    # == Params
    #
    # === Required
    # [environment_bindings (Hash|String|URI String)]  The environment_bindings argument will accept
    #                                                  a hash object, a JSON formatted String, or a 
    #                                                  URI to a local or remote JSON file. The contents
    #                                                  of this file will represent the settings required
    #                                                  to deploy an application to an environment.
    #
    def initialize(environment_bindings)
      
      if environment_bindings.is_a?(Hash)
        @bindings = environment_bindings
      elsif (File.exists? environment_bindings)
        json = File.open(environment_bindings, "r").read
        @bindings = create_hash_from_json(json)
      else
        @bindings = create_hash_from_json(environment_bindings)
      end
      
    end
    
    ##########
    # Accessor
    def devops_server_interface=(devops_server_interface)
      @devops_server_interface = devops_server_interface
    end
    
    ##########
    # Accessor so we can lazy initialize
    def devops_server_interface

      # Lazy Initialize @devops_server_interface
      if(@devops_server_interface.nil?)
        if(ENV["PROPS_FILE"])
          if(File.exists?(ENV["PROPS_FILE"]))
            props = YAML::load_file(ENV["PROPS_FILE"])
            env_def = bindings['ENVIRONMENT_DEF']
            if(props['devops_service_properties'])
              devops_svr_url = props['devops_service_properties']['url']
            end
            if(props['rtc_properties'])
              rtc_url = props['rtc_properties']['url']
            end
            if(props['iwd_properties'])
              iwd_url = "https://#{props['iwd_properties']['hostname']}"
            end
            log = Logger.new(props["log_file"] || STDOUT)
            log.level = Logger.const_get( (props["log_level"] || "INFO").upcase )
            @devops_server_interface = DevopsDeploymentLib::DevopsServerInterface.new(env_def, devops_svr_url, rtc_url, iwd_url, log)
          else # file doesn't exist
            puts("ENV[\"PROPS_FILE\"] is set to #{ENV["PROPS_FILE"]} which does not exist.  This prevents contacting the DevOps server")
          end
        else # PROPS_FILE not in the ENV
          puts("ENV[\"PROPS_FILE\"] is not set which prevents contacting the DevOps server")
        end
      end # end lazy init
      
      return @devops_server_interface
    end
    
    ##########
    # Pass through to the devops server interface
    def post(build_id, build_uuid, application_status, infrastructure_status)
      if(devops_server_interface())
        return devops_server_interface().post(build_id, build_uuid, application_status, infrastructure_status)
      end
    end
    
    ##########
    # Pass through to the devops server interface
    def view_url
      if(devops_server_interface())
        return devops_server_interface().view_url
      end
    end
    
    ##########
    # Pass through to the devops server interface
    def cloud_resource_view_url
      if(devops_server_interface())
        return devops_server_interface().cloud_resource_view_url
      end
    end
    
    ##########
    # Pass through to the devops server interface
    def posted?
      if(devops_server_interface())
        return devops_server_interface().posted?
      end
    end
    
    ##########
    # Pass through to the devops server interface
    def update(cloud_resource_id, application_status, infrastructure_status)
      if(devops_server_interface())
        return devops_server_interface().update(cloud_resource_id, application_status, infrastructure_status)
      end
    end
    
    
    def get_binding_obj
      bh_obj = BindingHelper.new(@bindings)
      return bh_obj.get_binding()
    end
    
    # This internal class helps with generating an object
    # that has all of the Environment bindings attached to
    # class attributes. This is done because the ERB template
    # API requires a Binding object to determine the context
    # in which it looks for the values of the template variables
    class BindingHelper
      def initialize(hash_obj)
        hash_obj.each do |k, v|
          self.instance_variable_set("@#{k}".to_sym,v)
        end
      end
      
      def get_binding
        return binding()
      end
    end
  end
end
