require 'zk'
require 'thor'
require 'thor/actions'
require 'thor/util'
require 'thor/task'
require 'thor/runner'
require 'thor/group'
require 'weaver/system'
require 'weaver/utils/project_helper'
require 'weaver/coordination/system'

module Weaver
  module Cli
    
    # Base class for all Weaver CLI plugins. Derive from that
    # class and place your plugin code either:
    #  - the plugins/ folder in the this project
    #  - $HOME/.weaver/plugins/cli
    #  - in any GEM or on the LOAD_PATH when it is in weaver/plugins/cli
    #
    # Each plugin has to follow the Thor standards of implementing a command.
    # Please refer to the Thor documentation for more info.
    class WeaverCliPlugin < Thor
      include Weaver::Core

      class_option :verbose, :type => :boolean, :aliases => "-v", :desc => "Enable verbose logging"
      class_option :no_color, :type => :boolean, :desc => "Turn off colorized output"
      class_option :sysproperties, :type => :hash, :desc => "Set global properties made available to the Weaver runtime"
      class << self
        def plugins
          @plugins ||= []
        end

        def namespace_description(description)
          @description = description
        end

        def inherited(subclass)
          plugins << subclass
        end

        def get_description
          @description
        end

        # Scrape the Task's +usage+ to determine the argument names, then construct definitions for them
        def thor_args(task)
          retval = ''
          task.usage.split.drop(1).each { |param_name| retval.concat(" #{param_name}\t# #{self::ARG_DEFS[param_name]}\n") }
          return retval
        end

        # @see #task_help
        def banner(task, namespace = true, subcommand = true)
          if @banner_with_args && defined? self::ARG_DEFS
            # This is the same as the below, plus an additional 'Arguments' section
            "#{basename} #{task.formatted_usage(self, true, subcommand)}\n\nArguments:\n#{thor_args(task)}"
          else
            # This is a necessary fix otherwise the subcommand banner is displayed incorrectly (FR)
            "#{basename} #{task.formatted_usage(self, true, subcommand)}"
          end
        end

        # Invoke Thor::task_help with a customized banner that includes Weaver arguments.
        # It would be better to replicate Thor::task_help() instead of using this technique.
        def task_help(shell, task_name)
          @banner_with_args = true
          super
          @banner_with_args = false
        end
      end

      # All helpers need to go in the "no_tasks" wrapper otherwise Thor issues a warning
      # because it tries to treat it as new command line method.
      no_tasks do
        
      # Setups logging according to the system properties and verbose flag
      def setup_logging
        log_level = "fatal" # default
        # set to value from system properties if applied
        log_level = Weaver::System.instance.get_property(:log_level) unless Weaver::System.instance.get_property(:log_level).nil?
        # check if system  property set that is not yet applied (usually not)
        if options[:sysproperties]
          lvl = options[:sysproperties].select{ |key, value| key.to_s == "log_level" }[log_level]
          log_level = lvl unless lvl.nil?
        end
        # allow --verbose flag to override
        if options[:verbose]
          log_level = "debug"
        end
        
        # set logger level
        level = {"debug" => Logger::DEBUG, "info" => Logger::INFO, "warn" => Logger::WARN, "error" => Logger::ERROR, "fatal" => Logger::FATAL}[log_level.downcase]
        # backup in case 
        level = Logger::FATAL if level.nil?
        @@logger.level = level
        set_level = {Logger::DEBUG => "debug", Logger::INFO => "info", Logger::WARN => "warn", Logger::ERROR => "error", Logger::FATAL => "fatal"}[@@logger.level]
        @@logger.info %Q{Weaver logger level set to: #{set_level}}
        # make sure it is set so that coordination will know what to use
        Weaver::System.instance.set_property(:log_level, set_level)
      end
      
        # Collects all the system properties that are specified as CLI arguments
        def collect_properties
          if options[:sysproperties]
            options[:sysproperties].each do |key, value|
              Weaver::System.instance.set_property(key, value)
              setval = Weaver::System.instance.get_property(key)
              @@logger.debug "Setting Weaver system property '#{key}' to '#{setval}'"
            end
          end
        end
        
        # Report error and exit unless +file+ exists
        def validate_file_exist?(file)
          unless File.exists?(file)
            say_status("Error:", "File '#{file}' does not exist.", :red)
            exit
          end
        end        
        
        # Uses the ProjectHelper class to load a project
        def project_loader(directory, config_overrides, config_filename_override = nil)
          begin
            Weaver::Utils::ProjectHelper.new(directory, config_overrides, config_filename_override)
          rescue Weaver::Core::WeaverError => ex
            say_status("Error", ex.message, :red)
            exit 1
          end         
        end 
        
        # Create a ZooKeeper coordinator client
        def create_coordinator
          coordinator_hosts = Weaver::System.instance.get_property(:zookeeper_hosts)
          if coordinator_hosts.nil? # in case someone forgot to call init first - and we don't want to call it twice
            init()
            coordinator_hosts = Weaver::System.instance.get_property(:zookeeper_hosts)  # re-read after init                      
          end
          @@logger.debug "Coordination hosts are: #{coordinator_hosts.inspect}"
          Weaver::Coordination::System.instance.set_property(:log_level, Weaver::System.instance.get_property(:log_level))
          Weaver::Coordination::System.instance.set_property(:zookeeper_hosts, coordinator_hosts)
          Weaver::Coordination::System.instance.set_property(:zookeeper_port, (Weaver::System.instance.get_property(:zookeeper_port)|| 2181))
        end
        
        # inits a project by setting up logging, collects system properties and loads the project properties from the current directory
        def init (config_overrides = {})
          # set initial log level to INFO while load properties, then set it per the properties
          @@logger.level = Logger::INFO
          project = project_loader(Dir.pwd, config_overrides) if config_overrides
          collect_properties # command line properties override project configuration properties
          setup_logging
          return project if config_overrides
          return nil
        end

        # Parses the deployment parameters and transforms them into a Hash
        def parse_deploy_parameters(deploy_param_string)
          deploy_params = Hash.new

          # parsing cli_params and convert key to symbol
          unless deploy_param_string.nil?
            say_status("Warning", "--deploy-params option is deprecated", :yellow)
            
            list = deploy_param_string.split(' ')
            list.each do |item|
              if item.include?('=')
                value = item.split('=')
                deploy_params[value[0]] = value[1]
              else
                say_status("Error", "Unable to parse 'deploy_params'. No separator '=' found in '#{item}', deploy_params='#{options[:deploy_params]}'.", :red)
                exit 1
              end
            end
          end
          deploy_params.symbolize
        end

      end

    end

  end
end