require 'weaver/core/operating_system'
require 'weaver/core/os/os_type'

module Weaver
  module Core
    # \Weaver::Core::Os provides parsing of the +os+ statement on a Node.
    #
    # @see OperatingSystemHandler
    # @see OsType
    module Os
      # Author:: Florian Rosenberg
      class OperatingSystemHandler
        include Weaver::Core

        # Contains a nested map of known operating systems. The know  
        # 
        attr_accessor :known_oses
        
        # Creates a new OperatingSystemHandler
        def initialize()
          @known_oses = Hash.new
          load()          
        end

        # Matches an OS based on a given os_string by using a regular expression. 
        # The known operating systems are defined in the files <os_type>_os.rb in 
        # folder lib/weaver/core/os using a simple DSL.   
        #
        # * *Args*    :
        #   - +os_string+ -> The unstructured OS string.
        # * *Returns* :
        #   - A map of properties about the OS that corresponds to the keys defined
        #     in the <os_type>_os.rb files. 
        #
        def match_os(os_string)
          @@logger.debug("OS String to match: #{os_string}")
          @known_oses.each { |k, v|            
            regex = v[:regex]
            match = regex.match(os_string) unless regex.nil?
            if !match.nil?
              # In Ruby 1.8.x MatchData.[] did not take a name the way it does in 1.9.x.
              # As we only have one capture we will use 0 instead of the name.
#              if match[:version].nil?
              if match[0].nil?
                return v
              else
                retval = v.clone()
                # In Ruby 1.8.x MatchData.[] did not take a name the way it does in 1.9.x.
                # As we only have one capture we will use 0 instead of the name.
                retval[:version] = match[0]
                return retval
              end
            end
#            return v unless match.nil?
          }
          return nil
        end
        
        def match_architecture(arch_string)
          match = :unknown
          if /(.*x86.*)?.*32.*(bit)*.*/i.match(arch_string)
            match = :x86_32
          elsif /(.*x86.*)?.*64.*(bit)*.*/i.match(arch_string)
            match = :x86_64
          end
          return match
        end      
      

        # Loads all custom OS and distribution definitions
        protected

        def load()
          @@logger.debug "Loading " + File.dirname(__FILE__)
          Dir[File.dirname(__FILE__) + "/*_os.rb"].each do |file|
            OsType.new(self).instance_eval(File.read(file))
          end
        end

      end
    end

  end
end