require 'weaver/core/logging.rb'
require 'weaver/core/base.rb'
require 'weaver/validation/validation_error'
require 'weaver/core/resolution'
require 'weaver/iwd/pattern_utils'
require 'weaver/core/status_constants'

module Weaver
  module Core
    # Represents a Weaver +rule+ model element.
    #
    # Author:: Ed Snible
    class Rule
      include Weaver::Validation
      
      # Weaver Symbol which acts as the internal id of the rule
      attr_accessor :id

      # The scope of the rule, a Proc which returns a Base or Array of Base to which this rule's :body applies.
      attr_accessor :where

      # \String which acts as the public id of the rule
      attr_accessor :rule_id

      # Textual description of the rule
      attr_accessor :description

      # The validation phase for checking the Rule; +:predeployment+ or +nil+ are the usual values
      attr_accessor :phase

      # The tags associated with the rule
      attr_accessor :tags

      # Array of Resolutions associated with the rule
      attr_accessor :resolutions

      # +lambda+ holds Ruby code to validate one Weaver entity returned by this Rule's #where clause.
      # The body receives a single parameter, the entity to be checked, and calls #report if the passed
      # entity manifests the problem the rule checks for.
      attr_accessor :body

      # \String that tells .weaver source file and line number containing this Rule's definition.
      attr_accessor :weaver_origin
      
      # Instance of \ContinuousValidation containing the metadata corresponding to the continuous validation block of this rule. 
      attr_accessor :contval
      
      # * +id+ - Weaver ID for the rule
      def initialize(id)
        @id = id
        @resolutions = []
      end

      # :stopdoc:
      def to_s
        return "Rule #{super.to_s}"
      end
      # :startdoc:

      # The number of matches this rule has in the supplied Environment
      # * +env+ - Environment
      def match_count(env)
        return 0 if where.nil?

        sandbox = ValidationSandbox.new(env)
        where_matches = sandbox.instance_eval &@where
        return 0 if where_matches.nil?
        return 1 unless where_matches.respond_to?(:size)
        return where_matches.size
      end

      # Run the rule over +root+ and append problems discovered to +out_messages+.
      # * +root+ - Weaver model to check
      # * +out_messages+ - Array of Weaver::Validation::ProblemStatus
      # * +desired_phases+ - Array of Symbol, the Symbols allowed include +:predeployment+
      # * +excluded_tags+ - 
      def validate(root, out_messages, desired_phases = nil, excluded_tags = nil)
        @@logger.debug "Evaluating rule :#{id}"
        # TODO restore the next line
        #raise WeaverError.new "no where clause for rule at #{@weaver_origin}" if where.nil?
        return if where.nil?

        #puts "EX = #{excluded_tags.inspect} -- size = #{excluded_tags.length}"

        if (!excluded_tags.nil?) && excluded_tags.length > 0 && (!@tags.nil?) && @tags.length > 0
          ## User wants to exclude rules with certain tags
          ## This rule has been associated with at least one tag
          if  (excluded_tags & @tags) != []
            # This rule is associated with a non-desired tag
            @@logger.debug "Skipping rule :#{@id}. Rule tags = #{@tags.inspect}. Excluded tags = #{excluded_tags.inspect}"
            return
          end
        end

        if (desired_phases.nil? || desired_phases.length == 0)
          ## User wants universal rules only
          if !@phase.nil? || @phase == ""
            # This rule is not universal, i.e., it has a phase; thus, it needs to be skipped.
            @@logger.debug "Skipping rule :#{@id}. Rule phase = #{@phase}. User wants universal rules only."
            return
          end
        elsif (!@phase.nil?) && @phase != "" && (!desired_phases.member?(@phase))
          ## This rule is not universal and does not belong to the desired phases. Skip it.
          @@logger.debug "Skipping rule :#{@id}. Rule phase = #{@phase}. Desired phases = #{desired_phases.inspect}"
          return
        end

        sandbox = ValidationSandbox.new(root)
        where_matches = sandbox.instance_eval &@where

        # If there where{} returned only a single object wrap it in an Array
        unless where_matches.nil? || where_matches.respond_to?(:size)
          where_matches = Array.[](where_matches)
        end

        @@logger.debug "  (#{where_matches.nil? ? 0 : where_matches.size} 'where' clause matches)"
        unless where_matches.nil?
          where_matches.each do |match|
            @unreported_messages = []
            begin
              body.call(match)
              out_messages.concat(@unreported_messages)
              @unreported_messages = []
              #              @@logger.debug "Succeeded eval"
            rescue Weaver::Validation::ValidationError => ve
              raise ve                    # Rethrow validation errors intended by programmer
            rescue Exception => e         # Unexpected exceptions get logged
              context = where.kind_of?(Base) ? context : Topology.new(:dummy, nil)
              out_messages.push ProblemStatus.new(Logger::ERROR, "Rule :#{id} body threw exception #{e}", context, weaver_origin, "rule failure", e)
              @@logger.debug "rescued #{e} with backtrace #{e.backtrace}"
            end
          end
        end
      end

      # \Rule bodies may call this method to report problems discovered
      def report(level, message, context, exception = nil, rule_type=rule_id, rule_description=description, rule_resolutions = [])
        raise WeaverError.new("'context' should be a Base but it is a #{context.class} at #{caller}") unless context.kind_of?(Base)

        if (rule_type.nil?)
          rule_type = rule_id
        end
        if (rule_description.nil?)
          rule_description = description
        end
        all_resolutions = resolutions.dup
        rule_resolutions.each { | res |
          all_resolutions << res
        }

        problem_status = Weaver::Validation::ProblemStatus.new(level, message, context, nil,
        rule_type, exception, rule_description, Weaver::Iwd::PatternUtils.evaluated_resolutions(all_resolutions))
        @unreported_messages.push(problem_status)
        return problem_status
      end

      # * +label+ - label for a manual or automated procedure to fix the problem detected by the rule
      # * +description+ longer description for the problem detected by the rule
      # * +type+
      # * +fix_proc+ optional code to change the model to fix the problem
      def add_resolution(label, description=nil, type=Weaver::Core::ResolutionTypes::DEFAULT, fix_proc=nil)
        @resolutions << Resolution.new(label, description, type, fix_proc)
      end
      
      # Return the model elements to which the rule applies.
      # * +root+ - Weaver model to check
      # * +out_messages+ - Array of Weaver::Validation::ProblemStatus
      def compute_targets(root, out_messages)
        @@logger.debug "Computing targets of rule :#{id}"
        return if where.nil?
        sandbox = ValidationSandbox.new(root)
        where_matches = sandbox.instance_eval &@where

        # If there where{} returned only a single object wrap it in an Array
        unless where_matches.nil? || where_matches.respond_to?(:size)
          where_matches = Array.[](where_matches)
        end
        
        return where_matches
      end

    end

    protected
    
    # a 'clean room' pattern to evaluate Rule +where+ clauses
    # Author:: Ed Snible
    class ValidationSandbox
      # Root of the Weaver document to be checked
      attr_reader :root
      
      # * +root+ Root of the Weaver document to be checked
      def initialize(root)
        raise WeaverError.new("Rule now expects Root") unless root.kind_of?(Root)

        # Typically 'root' will contain either an environment or an infrastructure but not
        # both.  However some rules a non-nil environment or a non-nil infrastructure so the
        # sandbox will provide a bogus one if needed.

        @root = root

        if root.env.nil? && root.infrastructure.nil?
          # We were supplied neither
          @unfolded_environment = Topology.new(:dummy, nil)
          @unfolded_infrastructure = Topology.new(:dummy, @unfolded_environment)
          @unfolded_environment.topologies.merge!(root.topologies)
        elsif root.env.nil?
          # We were supplied an infrastructure but no environment
          @unfolded_environment = Topology.new(:dummy, nil)
          @unfolded_infrastructure = root.infrastructure
          @unfolded_environment.topologies.merge!(root.topologies)
        else
          # We were supplied an environment, possibly without infrastructure
          @unfolded_environment = root.env
          infrastructure = root.env.topologies.values.find { |top| top.weaver_instance_of?(:infrastructure) || top.cloud != nil }
          @unfolded_infrastructure = infrastructure.nil? ? Topology.new(:dummy, root.env) : infrastructure
        end
      end

      # #root's folded environment Topology
      def folded_environment
        if @environment.nil?
          @environment = Weaver::Transforms::Fold.fold(environment)
          @infrastructure = @environment.topologies.values.find { |top| top.weaver_instance_of?(:infrastructure) }
        end
        return @environment
      end

      # #root's folded infrastructure Topology
      def folded_infrastructure
        folded_environment
        return @infrastructure
      end

      # #root's unfolded environment Topology
      def environment
        return @unfolded_environment
      end

      # #root's unfolded infrastructure Topology
      def infrastructure
        return @unfolded_infrastructure
      end

    end
  end
end