require 'weaver/core/base.rb'
require 'weaver/deployers/service_sandbox'

module Weaver
  module Core
    # Represents a service type on a Topology, Node, or Component
    # @see ServiceStatement
    #
    # Author:: Ed Snible
    # Copyright:: IBM
    class Service < Base
      include Weaver::Core::Generation

      # Deployment code
      attr_accessor :predeploy_block
      attr_accessor :deploy_block
      attr_accessor :postdeploy_block

      # Deployment dataflow model
      attr_accessor :predeploy_dataflow
      attr_accessor :deploy_dataflow
      attr_accessor :postdeploy_dataflow

      # Undeployment code
      attr_accessor :undeploy_block

      # Undeployment dataflow model
      attr_accessor :undeploy_dataflow
      
      # * +id+ Symbol to represent the entity
      # * +parent+ The enclosing Base
      def initialize(id, parent)
        super(id, parent)
        @type = :service
        @predeploy_dataflow = {}
        @deploy_dataflow = {}
        @postdeploy_dataflow = {}
        @undeploy_dataflow = {}
        @thread = nil
      end

      # Validates the entity and its direct and indirect contained children.
      # * +out_messages+ ProblemStatus will be pushed onto this for any validation message
      # * +desired_phases+ optional Array which may be +[:predefined]:
      def validate_collect(out_messages, desired_phases = nil)
        @@logger.debug "Validating service '#{fully_qualified_id}'"
        
        super
        
        [predeploy_dataflow, deploy_dataflow, postdeploy_dataflow, undeploy_dataflow].each do |dataflow_hash| 
          unexpected_dataflow_categories(dataflow_hash).each do |category|
            out_messages.push ProblemStatus.new(Logger::WARN, "Unknown dataflow category #{category.inspect}", self)
          end
        end
      end
      
      def predeploy(deploy_parameters)
        raise WeaverError, "Service thread object not nil" unless @thread.nil?
        unless predeploy_block.nil?
          @thread = Thread.new do
            sandbox = Weaver::Deployers::ServiceSandbox.new(self, 'predeploy', predeploy_dataflow)
            sandbox.instance_exec(deploy_parameters, &predeploy_block)
          end
        end
      end
      
      def deploy(deploy_parameters)
        raise WeaverError, "Service thread object not nil" unless @thread.nil?
        unless deploy_block.nil?
          @thread = Thread.new do
            sandbox = Weaver::Deployers::ServiceSandbox.new(self, 'deploy', deploy_dataflow)
            sandbox.instance_exec(deploy_parameters, &deploy_block)
          end
        end
      end
      
      def postdeploy(deploy_parameters)
        raise WeaverError, "Service thread object not nil" unless @thread.nil?
        unless postdeploy_block.nil?
          @thread = Thread.new do
            sandbox = Weaver::Deployers::ServiceSandbox.new(self, 'postdeploy', postdeploy_dataflow)
            sandbox.instance_exec(deploy_parameters, &postdeploy_block)
          end
        end
      end
      
      def undeploy
        raise WeaverError, "Service thread object not nil" unless @thread.nil?
        unless undeploy_block.nil?
          @thread = Thread.new do
            sandbox = Weaver::Deployers::ServiceSandbox.new(self, 'undeploy', undeploy_dataflow)
            sandbox.instance_exec(&undeploy_block)
          end
        end
      end
      
      def join
        unless @thread.nil?
          @thread.join
          @thread = nil
        end
      end
      
      private
      
      def unexpected_dataflow_categories(hsh)
        hsh.keys.reject { |key| key == :inputs || key == :outputs }
      end
    end
  end
end