require 'weaver/system' 
require 'weaver/transforms/duplicate'
require 'weaver/core/logging'
require 'weaver/core/attribute'
require 'weaver/core/automation'
require 'weaver/core/component'
require 'weaver/core/node'
require 'weaver/transforms/unroll'

module Weaver
  module Deployers
    module Execution
        
    # Registers with the coordinator and provides dependency details
    #
    class CoordinationUtil
      include Weaver::Core
      
      # properties on Weaver::Core::Node that should always be lookuped via coordinator
      ALWAYS_LOOKUP = ['ip_address', 'hostname']
      WEAVER_COORDINATION_ROLE = "weaver_coordination"
        
      # Insert the coordination automation into a node. It is the first automation to be executed. 
      #
      # *Args*    :
      # * +node+ node
      # * +deploy_parameters+ External attributes that need to be set. In particular, the following are expected:
      # * ++:weaver_coordination_topology_id++ Identifier of the deployed environment instance
      # * ++:weaver_coordination_dependencies++ Hash of dependencies in the topology
      # *Impact* : Updates node model with new automation; if expected deploy parameters are missing, no change is made
      #
      def self.add_coordination_automation (node, deploy_parameters = {})
        coordination_id = deploy_parameters[:weaver_coordination_topology_id]
        dependencies = deploy_parameters[:weaver_coordination_dependencies]
        coordinator = deploy_parameters[:coordinator]
        
        # check if the coordinator is a real coordinator; also set coordinator_hosts
        coordinator_hosts = []
        if coordinator.nil?
          # TODO this is legacy to be removed
          coordinator_hosts = Weaver::System.instance.get_property(:zookeeper_hosts) if coordinator.nil?
          return nil if coordinator_hosts.empty?
        else
          coordinator_hosts = coordinator.hosts
          return nil unless coordinator.can_coordinate?
        end 
        # check other inputs too
        return nil if coordination_id.empty? or dependencies.nil?
              
        # We don't want to explicitly import this:
        
        #chef_automation (:weaver_coord) {
        #    role "weaver_coord"
        #    rolefile "/tmp/weaver_automation/roles/weaver_coord.rb"
        #    property :build_essential_compiletime, :name => "build_essential.compiletime", :required => false
        #    property :chef_handler_handler_path, :name => "chef_handler.handler_path", :required => false
        #    property :chef_handler_root_user => "root", :name => "chef_handler.root_user", :required => false, :validators => [value_type(:string)]
        #    property :chef_handler_root_group => "root", :name => "chef_handler.root_group", :required => false, :validators => [value_type(:string)]
        #    property :weaver_coordination_coordinator_hosts, :name => "weaver_coordination.coordinator_hosts", :required => false
        #    property :weaver_coordination_coordinator_port => "2181", :name => "weaver_coordination.coordinator_port", :required => false, :validators => [value_type(:string)]
        #    property :weaver_coordination_topology_id, :name => "weaver_coordination.topology_id", :required => false, :validators => [value_type(:string)]
        #    property :weaver_coordination_node_id, :name => "weaver_coordination.node_id", :required => false, :validators => [value_type(:string)]
        #    property :weaver_coordination_dependencies => {}, :name => "weaver_coordination.dependencies", :required => false
        #}    

        # And then set these:
        
        # node.weaver_coord.build_essential_compiletime = true
        # node.weaver_coord.weaver_coordination_node_id = <node_id>
        
        # So we use this programmatic approach instead:
        automation_id = :weaver_coordination
        
        automation = node.root_entity.prototype_instance(:chef_automation, automation_id, node, 'coordination_util.rb')

        automation.role = WEAVER_COORDINATION_ROLE
        automation.deps_file_url = deploy_parameters[:deps_file_url]
        
        add_attribute(automation, "build_essential", "compiletime", true)
        add_attribute(automation, "chef_handler", "handler_path")
        add_attribute(automation, "chef_handler", "root_user", "root")
        add_attribute(automation, "chef_handler", "root_group", "root")
        add_attribute(automation, "weaver.coordination", "coordinator_hosts", coordinator_hosts)
        add_attribute(automation, "weaver.coordination", "coordinator_port", 2181)
        add_attribute(automation, "weaver.coordination", "coordination_id", coordination_id)
        add_attribute(automation, "weaver.coordination", "node_id", node.id.to_s)
        add_attribute(automation, "weaver.coordination", "fq_node_id", node.fully_qualified_id)
        add_attribute(automation, "weaver.coordination", "read_dependencies", get_read_dependencies(node))
        add_attribute(automation, "weaver.coordination", "write_dependencies", get_write_dependencies(node, dependencies))
        add_attribute(automation, "weaver.coordination", "role_dependencies", role_dependencies(node))
        # TODO: this really depends on the number of sequential deploys...
        timeout = node.root_entity.config[:timeout]
        add_attribute(automation, "weaver.coordination", "timeout", timeout)
        
        node.automation[automation.id] = automation
        node.automation_order.insert(0, automation)
        
        return automation
      end # def self.add_automation
      
      # Identify write dependencies for a node. When such attributes are written at runtime, the value should
      # also be written to the runtime coordinator.
      def self.get_write_dependencies(node, dependencies = nil)
        if dependencies.nil?
          root = node.root_entity
          dependencies = get_dependencies(root)
        end
        return dependencies.select { |d| d[:source].scope_of_type(Weaver::Core::Node) == node }.inject(Set.new()) do |set, dependency|
          set << automation_keys(dependency[:source])
          set
        end.to_a
      end # def get_write_dependencies
      
      # Identify read dependencies for a node. When such attributes are read at runtime, they will need 
      # to be computed with input (calculated here) and the coordinator.
      def self.get_read_dependencies(node)
        return read_dependencies(node).map do |dependency_record|
          read_dependency = automation_keys(dependency_record[:target])
          read_dependency.merge!(:late_binding_expr => dependency_record[:late_binding_expr], 
                                :values => dependency_record[:values], 
                                :mapping => dependency_record[:mapping])
          read_dependency
        end # read_dependencies(node).map do
      end # def get_read_dependencies
      
      def self.get_dependencies (env)
        propagations = Weaver::Service::PropagationChecker.model_propagations(env)
        
        # This produces a Hash:  Weaver::Core::Automation (automation) --> Hash
        # The second hash is String(property_id) --> Set<Attribute>
        
        # We should think of this as list of:  [automation, property_id] depends_on Set<Attibute>
        # This method turns it into a different Hash containing only id's that will be useful to the
        # ChefInterceptor running on a VM.  See below for details.

        # nodes = []
        dependencies = []
        propagations.each do |target_automation, properties|
          properties.each do |target_property_id, depends_on|
            depends_on.each do |source_property, mapping|
              target = target_automation.amd[target_property_id]
              source = source_property
              @@logger.debug "Considering dependency: #{target.fully_qualified_id}  <== #{source.fully_qualified_id}"
              unless should_reject?(target, source) then #source_id.empty? or automation_id.first == source_id.first
                d = {:target => target, :source => source}
                @@logger.debug "Recording dependency: #{d.inspect}"
                dependencies << d
              end # unless should_reject?()
            end # depends_on.each do
          end # properties.each do
        end # propagations.each do
        
        # The above creates an Array of Hash where each elements is a pair of target, source Weaver::Core::Base objects
        @@logger.debug %Q{All dependencies are:\n #{dependencies.map { |d| "#{d[:target].fully_qualified_id} <== #{d[:source].fully_qualified_id}" }.join("\n")}}

        return dependencies        
      end # def self.get_dependencies
      
      # Records data about the environment in the coordinator including:
      #   - deployment id, status for environment
      #   - hostname, ip_address, status(es) of nodes
      def self.write_defaults (transaction, root)
        return if root.topologies.empty?
        
        [:deployment, :config_status].each do |attr_id|
          attr = root.topologies.values.first.amd[attr_id]
          transaction.write_data(attr.fully_qualified_id, attr.value) unless attr.nil? or attr.value.empty?
        end # [].each do
      
        root.flatten_children.select {|c| c.is_a?(Weaver::Core::Node) && !Weaver::Transforms::Unroll.template?(c) }.each do |node|
          # TODO Do we need to do this if it is already there?
          [:ip_address, :hostname, :config_status, :vm_status].each do |attr_id|
            attr = node.amd[attr_id]
            transaction.write_data(attr.fully_qualified_id, attr.value) unless attr.nil? or attr.value.nil? or attr.value.empty?
          end # [].each do
          # TODO elastic ip addresses, storage?
        end # @operation_map.each do
      end # def write_defaults
      
      # Write values to coordinator from the model. This should be done for attributes on which there are dependencies
      # defined (because they are cross node) but for which the value is available in the model.
      # *Args*
      #  + dependencies + dependencies that are locally resolvable; assumed to be so 
      ### DEPRECATED - should no longer be necessary; values are inserted in table of values sent to VM
      def self.write_constants (transaction, dependencies)
        locally_resolvable_dependencies = dependencies.select do |d| 
          get_container(d[:source], Weaver::Core::Automation).nil? and !['hostname', 'ip_address'].member?(d[:source].id)
        end
        locally_resolvable_dependencies.each do |dependency|
          source = dependency[:source]
          source_node = get_container(source, Weaver::Core::Node)
          # The path from the node to the attribute
          suffix = source.fully_qualified_id.sub(%r{^#{source_node.fully_qualified_id}}, '')
      
          # All the nodes that need to be considered (group members)
          nodes = Weaver::Transforms::Unroll.template?(source_node) ? Weaver::Transforms::Unroll.group(source_node).members.values : [source_node]
          nodes.each do |node|
            value = node.instance_eval(suffix.sub(/^\./, ''))
            transaction.write_data(node.fully_qualified_id + suffix, value.to_s) unless value.nil?
          end # nodes.each do
        end # locally_resolvable_dependencies.each do
      end # def self.write_constants
      
      # TODO: read from config
      def self.get_coordinator
        coordinator_hosts = Weaver::System.instance.get_property(:zookeeper_hosts)
        port = Weaver::System.instance.get_property(:zookeeper_port) || 2181
        @@logger.debug "Creating coordinator with servers #{coordinator_hosts} and port #{port}"
        coordinator = Weaver::Coordination::CoordinatedTransactionManager.instance.add_coordinator(coordinator_hosts, port)
        @@logger.debug "Created coordinator #{coordinator.inspect}"
        return coordinator
      end # def foo

      
      protected
      
      def self.add_attribute(automation, prefix,postfix,value=nil)
        attribute_id = (prefix+"."+postfix).gsub(/[^_a-zA-Z0-9]/,'_').to_sym
        attribute = Weaver::Core::Attribute.new(attribute_id, automation)
        attribute.name = (prefix+"."+postfix)
        attribute.value = value unless value.nil?
        attribute.required = false
        attribute.mutable = true
        automation.amd[attribute.id] = attribute
      end # def self.automation.add_attribute

      # Identify all the attributes (of Weaver::Core::Automations) on a node that have dependencies on 
      # other attributes.  In particular, those that have run time read dependencies.
      #
      # *Result*
      # Array<{:target => <target attribute>, :late_binding_expr => <sourcified late binding expression for target attribute>, :values => <Hash of source attribute fully qualified id => value>}>
      def self.read_dependencies (node)
        target_details = []
      #          propagations = Weaver::Service::PropagationChecker.model_propagations(node)
      #          propagations = @propagations.select { |target_automation, properties| equivalent_node?(node, target_automation.scope_of_type(Weaver::Core::Node)) }
        node.flatten_children.select { |child| child.is_a?(Weaver::Core::Automation) }.each do |target_automation|
        # propagations.each do |target_automation, properties|
          properties = Weaver::Service::PropagationChecker.model_script_propagations(target_automation)
          properties.each do |target_property_id, depends_on|
            lookup, details = requires_lookup?(depends_on, node)
            @@logger.debug "Property #{target_automation.fully_qualified_id}.#{target_property_id} requires lookup: #{lookup}"
            @@logger.debug "depends_on is:#{depends_on}"
            if lookup
              # At least one property requires lookup so:
              #   (a) Sourcify LateBinding for target_property for evaluation on the target
              #   (b) Insert values for all referenced (source) properties to values array
              # This array will be available for value lookup as an alternative to coordinator lookups  
              # (a)
              @@logger.debug "   attempting to sourcify #{target_automation.fully_qualified_id}.#{target_property_id}"
              @@logger.debug "   that is, #{target_automation.amd[target_property_id]}"
              @@logger.debug "   or, #{target_automation.amd[target_property_id].unresolved_value.class}:::#{target_automation.amd[target_property_id].unresolved_value}"
              source_expr = target_automation.amd[target_property_id].unresolved_value.to_source
              source_expr.gsub!(/^late_binding/, 'proc')
              source_expr.gsub!(/\n/, '__NEWLINE__')
              @@logger.debug "   sourcify returns: #{source_expr.inspect}"
              # (b)
              values = {}
              details.each do |source_property, info|
                values[source_property.fully_qualified_id] = info[:value]
                @@logger.debug "   added values[#{source_property.fully_qualified_id}] = #{values[source_property.fully_qualified_id]}"
              end # details.each do
              attr_mapping = get_attribute_mappings(depends_on)
              @@logger.debug "   mapping is #{attr_mapping.inspect}"
              target_details << {:target => target_automation.amd[target_property_id], :late_binding_expr => source_expr, :values => values, :mapping => attr_mapping}
            end # if lookup
          end # properties.each do
        end # @propagations.each do
        
        return target_details
      end # def read_dependencies
      
      # Create a Hash of role --> things it should wait for before execution
      # *Args*
      #  - the node
      # *Result*
      #  - Hash role name --> id of things it should wait for (via coordinator)
      def self.role_dependencies (node)
        # TODO filter by ... chef_automation and to be deployed...
        dependencies = {}
        node.automation.each do |id, automation|
          unless automation.depends_on.empty?
            # what it is a dependency on a group node?
            dependencies[automation.amd[:role].value] = dependency_references(automation.depends_on)
          end # unless automation.depends_on.empty?
        end # node.automation.each do
        
        unless node.depends_on.empty?
          dependencies[WEAVER_COORDINATION_ROLE] = dependency_references(node.depends_on)
        end # unless node.depends_on.empty?
        
        return dependencies
      end # def self.role_dependencies

      # TODO: this probably belongs somewhere else
      # Translate the list of dependencies into a list of references on which to wait.
      # This includes extending dependencies on groups into a dependency on each group member
      def self.dependency_references (dependencies)
        return dependencies.map { |dependency|
          if dependency.is_a?(Weaver::Core::Node)
            group = Weaver::Transforms::Unroll.group(dependency)
            if group.nil?
              dependency
            else
              group.members.values.select {|n| !Weaver::Transforms::Unroll.template?(n)}
            end # if group.nil?
          else
            dependency
          end # if dependency.is_a?(Weaver::Core::Node)
        }.flatten.map { |dependency| dependency.fully_qualified_id }
      end # def self.dependencies
      
      # Given a list of source properties used to compute a target property, determine if any of the
      # source properties require lookup via  a coordinator at execution time.
      # *Args*
      #   - +source_properties+ list of source properties (Weaver::Core::Attribute)
      #   - +target_node+ Weaver::Core::Node containing the target property
      # *Result* a pair items:
      #   - a boolean flag indicating whether or not at least one source property requires lookup
      #   - a Hash mapping the source properties to the results of pproperty_requires_lookup?
      def self.requires_lookup? (source_properties, target_node)
        details = {}
        lookup = false
        source_properties.keys.each do |source_property| 
          details[source_property] = property_requires_lookup?(source_property, target_node)
          @@logger.debug "   property_requires_lookup(#{source_property.fully_qualified_id}): #{details[source_property].inspect}"
          lookup = lookup || details[source_property][:lookup]
        end
        return lookup, details
      end
      
      # Determine if a property value should be looked up via the coordinator at runtime.
      # *Args*
      #   - +source_property+ source property (Weaver::Core::Attribute)
      #   - +target_node+ Weaver::Core::Node containing the target property
      # *Result* A Hash containing the following information:
      #   - +:lookup+ boolean indicating whether or not the value should be looked up at runtime
      #   - +:+value+ value if lookup at runtime is not required (nil if lookup is required)
      #   - +:type+ type of property
      # TODO We also want to return :type; how determine?
      def self.property_requires_lookup?(source_property, target_node)
        return {:lookup => false} if source_property.nil?
        return {:lookup => false} if !source_property.is_a?(Weaver::Core::Attribute) 
        
        return {:lookup => false, :value => source_property.value} if target_node.nil?
        source_node = source_property.scope_of_type(Weaver::Core::Node)
        return {:lookup => false, :value => source_property.value} if source_node.nil?
      
        source_automation = source_property.scope_of_type(Weaver::Core::Automation)
      
        # always require look up for some attributes, even if local          
        return {:lookup => true, :value => nil, :type => :string} if source_automation.nil? and ALWAYS_LOOKUP.include?(source_property.id.to_s)
          
        # never lookup local nodes (except for items in ALWAYS_LOOKUP, handled above)
        return {:lookup => false, :value => source_property.value} if equivalent_node?(source_node, target_node)
        
        # Node are not equivalent
          
        if source_automation.nil?
          # If source property is not from an automation, it is defined in the weaver document; use that value
          return {:lookup => false, :value => source_property.value}
        else
          return {:lookup => true, :value => nil, :type => :string}
        end
      end # def requires_lookup?
      
      # Given a hash of Weaver::Core::Attribute --> Set<alias>
      # Generate the mapping key --> fully qualified id of attribute
      def self.get_attribute_mappings (source_properties)
        retval = {}
        source_properties.each do |attr, mapping_set|
          @@logger.debug "inverting the mappings to #{attr.fully_qualified_id} from #{mapping_set.inspect}"
          mapping_set.each do |mapping|
            x = retval[mapping] || []
            x.push(attr.fully_qualified_id)
            retval[mapping] = x
          end # mapping_set.each do
        end # source_properties.each do
        @@logger.debug "returning #{retval.inspect}"
        return retval
      end # def get_attribute_mappings
      
      # Modify a dependency so that it is suitable for the weaver_coordination recipe; that is, 
      # as a Hash {:target => {:node, :operation, :parameter}, :source => {:node, :operation, :parameter}}
      ## DEPRECATED
      def self.dependency_for_weaver_automation (dependency)
        result = {}
        [:source, :target].each do |key|
          attribute = dependency[key]
          result[key] = automation_keys(attribute)
        end # [].each do
        
        @@logger.debug "Dependency as input to weaver automation is #{result.inspect}"
        return result
      end # def dependency_for_weaver_automation
      
      # Convert a single attribute to a triple of keys:  <node, operation, parameter>
      # In a chef_automation, this corresponds to <node, role, attribute>
      def self.automation_keys (attribute)
        automation = get_container(attribute, Weaver::Core::Automation)
        node = get_container((automation.nil? ? attribute : automation), Weaver::Core::Node)
        prefix = automation.nil? ? node.fully_qualified_id : automation.fully_qualified_id
        postfix = attribute.name.nil? ? attribute.id.to_s : attribute.name
      
        attribute_name = attribute.fully_qualified_id.sub(%r{^#{prefix}\.}, '')
        if /\./ =~ attribute_name
          # if contains '.' replace the end w/ postfix (name over id)
          attribute_name.sub!(/\.[^\.]*$/, ".#{postfix}")
        else
          # if does not contain '.', just use postfix (name over id)
          attribute_name = postfix
        end # if /\./ =~ attribute
      
        return {
          :node => node.id.to_s, 
          :operation => automation.nil? ? nil : (automation.amd[:role].nil? ? nil : automation.role.to_s), 
          :parameter => attribute_name
        }
      end # def 
      
      ## Suporting methods for get_dependencies
      
      def self.should_reject?(target, source = {})
        if target.nil? then
          @@logger.debug "Reject - invalid (nil) target"
          return true
        end
        # TODO other restrictions on the target?
        
        if source.nil? then
          @@logger.debug "Reject - invalid (nil) source"
          return true
        end
        # TODO other restrictions on the source?
        
        target_node = get_container(target, Weaver::Core::Node)
        source_node = get_container(source, Weaver::Core::Node)
        
        if target_node.nil?
          @@logger.debug "Reject - target not a node"
          return true
        end
        if source_node.nil?
          @@logger.debug "Reject - souce not a node"
          return true
        end
        
        if equivalent_node?(target_node, source_node) then
          # special case for ip_address and hostname
          if get_container(source, Weaver::Core::Automation).nil? and [:ip_address, :hostname].include?(source.id)
            @@logger.debug "Same node but ip_address or hostname exception"
            return false
          end # special case
          @@logger.debug "Reject - same node"
          return true
        end
        
        # TODO document this
        if [:multiplicity_index].member?(source.id) then
          @@logger.debug "Reject - multiplcity_index"
          return true
        end
        
        return false
      end # def self.should_reject?
      
      # DEPRECATED -- use Abstract#scope_of_type()
      def self.get_container(object, type)
        return object.scope_of_type(type)
#        until object.nil? or object.is_a?(type)
#          object = object.scope
#        end # until ...
#        return object
      end # def self.get_container
      
    # Nodes are the same if ids are the same
    # One is a group template and the other is a group members (of the same group) 
      def self.equivalent_node?(node1, node2)
        return false if node1.nil? and !node2.nil?
        return false if !node1.nil? and node2.nil?
        return true if node1.nil? and node2.nil?        
        return true if node1.fully_qualified_id == node2.fully_qualified_id
        
        return Weaver::Transforms::Unroll.group(node1) == Weaver::Transforms::Unroll.group(node2) if Weaver::Transforms::Unroll.template?(node1)
        return Weaver::Transforms::Unroll.group(node2) == Weaver::Transforms::Unroll.group(node1) if Weaver::Transforms::Unroll.template?(node2)
        
        return false
      end # def self.equivalent_node?
      
      end # class CoordinationUtil
    end # module Execution
  end # module Deployers
end # module Weaver