module Weaver
  module Dsl
    # Mixin AutomationContainer in statements that may contain {AutomationStatement}.
    #
    # Author:: Florian Rosenberg
    # Copyright:: IBM
    module AutomationContainer
      include Weaver::Core
      
      # * *Args* :
      #   -+base+ the AbstractStatement being mixed with
      def self.included(base)
        base.register_statement_handler([:script_handler])
      end

      # Invoked if we might be processing an automation or extended automation type
      # * *Args*    :
      #   -+sym_or_model+ any unknown keyboard, but we return unless it is a type of an Automation like :script_package or a model of an Automation
      #   -+id+ any value following the unknown keyword, but we return unless it is the id (Symbol or String) of an automation
      def script_handler(sym_or_model, id, &block)
        return nil unless is_a?(Weaver::Dsl::AutomationContainer) && (sym_or_model.kind_of?(Automation) || root_statement.model.weaver_extended_type?(sym_or_model, Automation))
        return nil unless id.kind_of?(Symbol) || id.kind_of?(String)
        
        prototype = (sym_or_model.kind_of?(Symbol) || sym_or_model.kind_of?(String)) ? root_statement.model.prototype_instance(sym_or_model, id, model, weaver_stackframe) : sym_or_model 
        automation(id, {}, prototype, &block)
      end
      
      # Create an Automation from an +automation+ type
      #
      # * *Args*    :
      #   -+id+ a Symbol or String.  If String, the executable defaults
      #   -+scriptdata+ Hash of extra metadata
      #   -+prototype+ optional prototype instead of :automation's default type
      #   -+block+ optional block of value overrides
      def automation(id, scriptdata = {}, prototype = nil, &block)
        @@logger.debug "automation[(#{prototype.id.to_s+') ' unless prototype.nil?}:#{id}]"
        
        return if fail_id_redefinition(id, Automation)
        
        automation_model = model.automation[id]
        if automation_model.nil?
          automation_model = prototype || root_statement.model.prototype_instance(:automation, id, model, weaver_stackframe)
          automation_model.executable = id if id.kind_of?(String) # If the ID is a String, default the executable 
          model.automation[id] = automation_model
        else
          root_statement.model.widen(automation_model, prototype) unless prototype.nil?
        end
        
        unless scriptdata.nil?
          legacy_params(automation_model, scriptdata[:parameters]) unless scriptdata[:parameters].nil?
        end
        
        unless block.nil?
          automation_statement = AutomationStatement.new(automation_model, self)
          
          begin
            automation_statement.instance_eval &block unless block.nil?
          rescue Exception => e
            @@logger.debug "Caught #{e.inspect} processing automation statement, backtrace=#{e.backtrace}"
            model.add_validation_failure("automation problem #{e.to_s}", weaver_stackframe(e.backtrace))
          end
        end
        
        return automation_model
      end
      
      # The language +automation_order+ and +script_order+ statements.
      #
      # See {+automation_order+}[link:STATEMENTS_rdoc.html#label-automation_order]
      # * *Args*    :
      #   - +order+ -> Array of <Array of Symbol>|<Symbol>
      def automation_order(*order)
        @@logger.debug "automation_order[#{order.inspect}]"
        @model.automation_order = order.map { |i| i.kind_of?(Array) ? i.map { |id| @model.automation[id] } : @model.automation[i] }           
        @model.automation_order.instance_variable_set(:@weaver_origin, weaver_stackframe)
      end
      alias script_order automation_order

      private
      
      def legacy_params(script, params)
        @@logger.warn "Legacy :parameters notation on #{model.type} :#{model.id}"
        params.each do |name,val|
          id = name.symbolize
          attr = Attribute.new(id, script)
          attr.name = name
          attr.value = val
          script.amd[id] = attr 
        end 
      end      

    end
  end
end