require 'weaver/core/logging'
require 'weaver/dsl/requires/type_handler'
require 'weaver/core/requirements'

module Weaver
  module Dsl
    module Requires
      # Author:: Ed Snible
      class ProductTypeHandler < TypeHandler

        def initialize()
          super [ :product, :products ]
        end

        def handle(parent, args)
          @@logger.debug "#{self.class} #{parent} #{args}"
#          return MemberRequirement.new(parent.model, args) 
          return ProductRequirement.new(parent.model, args) 
        end

      end
      
      # Author:: Ed Snible
      class ProductRequirement < Core::Requirement
        # requires_hash should have keys :product (referring to a Component.name) and
        # (optionally) :version referring to a Component :version
        def initialize(scope, requires_hash)
          super scope
          
          @required_product = requires_hash[:product]
          @required_version = requires_hash[:version]
          # TODO warn if we have any additional keys in requires_hash besides :description
        end
        
        # We are looking for "component (:x) { name "y" version 'z' }
        # The possible problems are that we have the wrong version, that we haven't modeled the
        # required product at all but it is there, or that the required product isn't there.
        # Another problem is multiple versions of the same product on the infrastructure. 
        def validate_collect(out_messages, desired_phases = nil)
          all_final_realizations.each do |inf_node|
            closest_match = [nil]
            if exact_match(inf_node.components.values, closest_match)
              return
            end
            if closest_match[0].nil?
              out_messages.push ProblemStatus.new(Logger::ERROR, generate_undefinedcomponent_message(inf_node), inf_node, weaver_origin)
            else
              out_messages.push ProblemStatus.new(Logger::ERROR, generate_wrongversion_message(inf_node, closest_match[0]), inf_node, weaver_origin)
            end
          end
        end
        
        # Returns true if there is an exact match
        # out_closest_match will be that match if true, otherwise it will be some other
        # component with the same @required_product
        def exact_match(components, out_closest_match)
          components.each do |component|
            if component.name == @required_product
              out_closest_match[0] = component
              if suitable_version(component.version)
                return true
              end
            end
          end
          return false
        end
        
        def generate_undefinedcomponent_message(target)
          return <<-eos
#{weaver_classname(@scope)} :#{@scope.id} at (#{@weaver_origin}) requires the
#{weaver_classname(target)} :#{target.id} at (#{target.weaver_origin}) to have
 
  component (:#{@required_product.symbolize}) {
    name '#{@required_product}'#{version_weaver}
  }
  
Configure the cloud image to meet this requirement and rediscover.  If discovery did not detect the
requirement but you are certain it is there you may add the component to the infrastructure node.

eos
        end
        
        def version_weaver
          if @required_version.nil?
            return nil
          else
            return "\n    version '#{@required_version}'"
          end
        end
        
        def generate_wrongversion_message(target, closest_match)
          return <<-eos
#{weaver_classname(@scope)} :#{@scope.id} at (#{@weaver_origin}) requires the
#{weaver_classname(target)} :#{target.id} at (#{target.weaver_origin}) to have version
#{@required_version} but it has version #{closest_match.version}
eos
        end
        
        def suitable_version(component_version)
          return true if @required_version.nil?
          
          if @required_version.kind_of?(Fixnum)
            @required_version = @required_version.to_s
          end
          
          if @required_version.kind_of?(String)
            return zephyr_suitable_version(component_version)
          elsif @required_version.kind_of?(Regexp)
            return @required_version.match(component_version) != nil
          elsif @required_version.kind_of?(Range)
            return @required_version.cover?(component_version.to_f)
          else
            raise WeaverError.new("Unsupported requirement :product range type #{@required_version.class}")
          end
          
        end
                
        # Is the version suitable, using the semantics of Rational Software Architect for Deployment Planning
        def zephyr_suitable_version(component_version)
          psections = @required_version.split('.')
          vsections = component_version.split('.')
          
          for i in 0..psections.length - 1
            # Does a further section end with '+'?
            lookaheadPlus = false
            for j in i+1 .. psections.length - 1
              if psections[j] =~ /\+$/
                lookaheadPlus = true
                break
              end
            end
      
            return true if lookaheadPlus && exceedsSection(psections[i], vsections[i])
            return false if !interpSection(psections[i], i < vsections.length ? vsections[i] : nil)
          end
      
          return true
        end
        
        private
        def interpSection(pattern, version)
          # pattern should be either 'x' or [digits]+ or [digits]+'+'
          return true if pattern == 'x'
          return comparePlus(pattern[0..pattern.length() - 1], version) if pattern =~ /\+$/
          return compare(pattern, version);
        end
        
        # Checks for versions definitely exceeding, for example "1..." doesn't exceed 1.1+ but "2" does.
        # return true if there is definite exceeding
        private 
        def exceedsSection(pattern, version)
          verConstraint = digits.to_i rescue -1
          return false if verConstraint < 0 

          # remove trailing letters
          version = stripTrailingNonDigits(version)

          return false if version.length == 0
          return version.to_i > verConstraint
        end
        
        private 
        def comparePlus(digits, version)
          return false if version.nil?

          verConstraint = digits.to_i rescue -1
          return false if verConstraint < 0

          # remove trailing letters
          version = stripTrailingNonDigits(version)
          return false if version.length == 0
          return version.to_i >= verConstraint
        end
        
        private
        def compare(digits, version)
          return false if version.nil?
          ver = digits.to_i rescue -1
          return false if ver < 0

          # remove trailing letters
          version = stripTrailingNonDigits(version)
          return false if version.length == 0

          return version.to_i == ver
        end
        
        private 
        def stripTrailingNonDigits(version)
          md = /[0-9]*/.match(version)
          retval = md.nil? ? "" : md[0]
          return retval
        end
        
      end
    end
  end
end