require 'weaver/core/logging'
require 'weaver/core/base'
require 'weaver/core/root'
require 'weaver/core/component'
require 'weaver/core/node'
require 'weaver/core/user'
require 'weaver/core/group'
require 'weaver/core/topology'
require 'weaver/core/automation'
require 'weaver/core/load_driver'
require 'weaver/core/policy'
require 'weaver/core/artifact'
require 'weaver/core/link'
require 'weaver/core/service'
require 'weaver/dsl/abstract_statement'
require 'weaver/dsl/import_container_statement'
require 'weaver/dsl/group_statement'
require 'weaver/dsl/user_statement'
require 'weaver/dsl/mixin_statement'
require 'weaver/dsl/load_driver_statement'
require 'weaver/dsl/policy_statement'
require 'weaver/dsl/provider_statement'
require 'weaver/dsl/nic_statement'
require 'weaver/dsl/automation_statement'
require 'weaver/dsl/topology_statement'
require 'weaver/dsl/service_statement'
require 'weaver/dsl/artifact_statement'
require 'weaver/dsl/base_statement'
require 'weaver/dsl/node_container'
require 'weaver/dsl/component_container'
require 'weaver/dsl/automation_container'
require 'weaver/dsl/link_container'
require 'weaver/dsl/load_driver_container'
require 'weaver/dsl/policy_container'
require 'weaver/dsl/artifact_container'
require 'weaver/dsl/user_container'
require 'weaver/dsl/group_container'
require 'weaver/dsl/topology_container_statement'

module Weaver
  
  module Dsl
    include Weaver::Core
    # The RootStatement represents the top-level context of a Weaver DSL.
    #
    # Author:: Ed Snible
    # Copyright:: IBM
    #
    # @see Root
    class RootStatement < AbstractStatement
      include ComponentContainer
      include NodeContainer
      include AutomationContainer
      include LinkContainer
      include LoadDriverContainer
      include PolicyContainer
      include GroupContainer
      include UserContainer
      include TopologyContainerStatement
      include ServiceContainerStatement
      include ProviderContainerStatement
      include ImportContainerStatement
      include ArtifactContainer

      # * +filename+ String filename
      # * +weaver_parser+ WeaverParser
      def initialize(filename, weaver_parser)
        @filename = filename
        @@logger.debug "root[filename: #{filename}]"
        super(filename, Root, nil, weaver_parser)
        @model = Root.new(filename, weaver_parser)
        
        @weaver_type_statements = {
          Component => ComponentStatement,
          Node => NodeStatement,
          User => UserStatement,
          Group => GroupStatement,
          Automation => AutomationStatement,
          LoadDriver => LoadDriverStatement,
          Policy => PolicyStatement,
          Link => LinkStatement,
          Topology => TopologyStatement,
          Import => ImportStatement,
          NetworkInterfaceController => NicStatement,
          Mixin => MixinStatement,
          Provider => ProviderStatement,
          Weaver::Core::Service => ServiceStatement,
          Base => BaseStatement,
          Artifact => ArtifactStatement
        }
      end
      
      # Process keyword 'type'
      def type(clazz, &block)
        # clazz can be a Hash for example {:db_component=>:component}, or just :db_component
        if clazz.kind_of?(Hash)
          clazzname = clazz.keys[0]
          basename = clazz.values[0]
          @@logger.debug "type[#{clazzname} < #{basename}]"
        else
          clazzname = clazz
          basename = nil
          @@logger.debug "type[#{clazzname}]"
        end
        
        augment_existing = model.weaver_extended_type?(clazzname, Base)

        begin
          class_statement = get_statement(clazzname, basename)
        rescue WeaverError => we
          model.add_validation_failure(we.message, weaver_stackframe(we.backtrace))
          @@logger.warn "Caught #{we.inspect} parsing; backtrace=#{we.backtrace}"
          return
        end
        
        class_statement.model.weaver_origin = weaver_stackframe
        class_statement.instance_eval &block unless block.nil?
        
        unless augment_existing
          # Convert attributes to implicit attributes
          class_statement.model.amd.each { |id, attr| attr.implicit = true }
           
          # Create a new class based on the template component
          model.make_new_class(class_statement.model)
        end
      end
  
      # Report an error (DEPRECATED)
      def report(*args)
        return @model.report(*args)
      end
      
      # Create Weaver DSL statement that can process the existing model entity (a Weaver::Core::Base)
      # * +entity+ any Weaver::Core::Base that is processed by the Weaver DSL
      # * +statement+ parent statement
      def statement_for_entity(entity, statement)
        class_statement_type = @weaver_type_statements[entity.class]
        if class_statement_type.nil?
          model.add_validation_failure("No DSL parser for class #{entity.class}")
          @@logger.warn "No DSL parser for class #{entity.class}; parsers are available for #{@weaver_type_statements.keys}, caller=#{caller}"
          return nil
        end
        class_statement = class_statement_type.new(entity, statement)
        class_statement.expression_context = true
        return class_statement
      end
      
      # Create Weaver DSL statement that can process the prototype given in the subtype definition      
      def get_statement(clazzname, basename)
        if model.weaver_extended_type?(clazzname, Base)
          raise WeaverError.new("#{clazzname} refinement cannot specify => symbol") unless basename.nil?
          
          # Allow user access to the actual prototype used
          prototype = model.prototype(clazzname)
        else
          if basename.nil?
            if clazzname.to_s =~ /component$/
              basename = :component
            elsif clazzname.to_s =~ /node$/
              basename = :node
            elsif clazzname.to_s =~ /user$/
              basename = :user
            elsif clazzname.to_s =~ /group$/
              basename = :group
            elsif clazzname.to_s =~ /load_driver$/
              basename = :load_driver
            elsif clazzname.to_s =~ /policy$/
              basename = :policy
            elsif clazzname.to_s =~ /infrastructure$/
              basename = :infrastructure
            elsif clazzname.to_s =~ /application$/
              basename = :application
            elsif clazzname.to_s =~ /automation$/
              basename = :automation
            elsif clazzname.to_s =~ /link$/
              basename = :link
            elsif clazzname.to_s =~ /mixin$/
              basename = :mixin
            elsif clazzname.to_s =~ /service$/
              basename = :service
            elsif clazzname.to_s =~ /artifact$/
              basename = :artifact
            end  
            
            if basename.nil?
              raise WeaverError.new("#{clazzname} needs => followed by one of #{model.weaver_extended_types.keys.inspect}")
            end
          end
          
          prototype = @model.prototype_instance(basename, clazzname, @model, weaver_stackframe)
        end
      
        class_statement_type = @weaver_type_statements[prototype.class]
        model.add_validation_failure("Unknown prototype class #{prototype.class}") if class_statement_type.nil?
        @@logger.warn "Unknown prototype class #{prototype.class}" if class_statement_type.nil?
        class_statement = class_statement_type.new(clazzname, self)
        class_statement.model = prototype
        class_statement.model.id = clazzname
        class_statement.model.type = clazzname
        class_statement.model.scope = nil
        return class_statement
      end
      
    end

  end # module Dsl
end # module weaver