require 'rexml/document'
require 'rexml/xpath'
require 'weaver/core/weaver_error'

module Weaver
  module Generators
    # The EclipseAppGenerator allows to generate a Weaver application definition
    # stub from set of Java projects in workspace directly or a specific workingset.
    class EclipseAppGenerator

      WORKINGSET_FILE = ".metadata/.plugins/org.eclipse.ui.workbench/workingsets.xml"

      attr_reader :workspace_location
      attr_reader :workingset_name
      # Creates a new EclipseAppGenerator with a specific Eclipse workingspace location
      # and an optional workingset file. If a workspace is given only, all projects in
      # the workspace are processes. If a workingset if given only the projects from
      # the workingset in the workspace are processed.
      # * *Args*    :
      #   - +workspace_location+ -> the location of the workspace to mine (String).
      #     If not given the application definition is not saved
      #   - +workingset_name+ -> the name of the workingset to process (String; optional)
      def initialize(workspace_location, workingset_name = nil)
        if File.exists?(workspace_location) &&
        File.directory?(workspace_location) &&
        File.readable?(workspace_location)
          @workspace_location = workspace_location
        else
          raise WeaverError.new("File '#{workspace_location}' is no directory, does not exist or is not readable.")
        end
        @workingset_name = workingset_name
      end

      # Generates and application definition and saves it to a file if given.
      # * *Args*    :
      #   - +application_id+ -> the id of the application
      #   - +filename+ -> the filename to store the file (String; optional).
      #     If not given the application definition is not saved
      #   - +mapping+ -> defines a node to component mapping (optional, Hash with id -> Array)
      #
      # * *Returns* :
      #   - the generated Application model
      #
      def generate_app_definition(application_id, filename, mapping = nil)
        app = Topology.new(application_id, nil)

        if @workingset_name == nil
          directories = Dir[File.join(@workspace_location, '/*')]
        else
          directories = find_relevant_directories()
        end

        if (mapping != nil)
          generate_app_definition_with_mapping(app, directories, mapping)
        else
          generate_app_definition_no_mapping(app, directories)
        end

        # write the app model to a file
        file = File.new(filename, 'w')
        file.write(app.to_weaver)
        file.close

        return app
      end

      protected

      def generate_app_definition_with_mapping(app, directories, mapping)
        # create nodes according to mapping
        mapping.each do |k, v|
          app.nodes[k] = Node.new(k, app)
        end

        directories.each do |dir|
          project_file = File.join(dir, '.project')
          classpath_file = File.join(dir, '.classpath')
          if File.exists?(project_file) # only process directory if .project file exists
            project_xml = File.new(project_file, 'r').read()

            # read the name of the project
            project_doc = REXML::Document.new(project_xml)
            project_name = REXML::XPath.match(project_doc, 'projectDescription/name/text()')[0].to_s
            project_sym = project_name.underscore.symbolize

            # find projects that it depends on
            dependencies = []
            project_doc.elements.each('projectDescription/projects/project') do |project|
              dependencies << project.text().to_s.underscore.symbolize
            end

            node_id = find_node_in_mapping(mapping, project_name)
            if node_id != nil
              # find all the above project depends on
              if File.exists?(classpath_file)
                classpath_xml = File.new(classpath_file, 'r').read()
                classpath_doc = REXML::Document.new(classpath_xml)
                classpath_doc.elements.each("classpath/classpathentry[@kind='src']") do |e|
                  path = e.attributes['path'].to_s.underscore
                  if path[0] == '/'
                    dependencies << path[1, path.length].to_s.symbolize
                  end
                end # block
              end

              # add component for it to node
              comp = Component.new(project_sym, app.nodes[node_id])
              comp.name = project_name
              app.nodes[node_id].components[project_sym] = comp

              # add a component requirement for each dependency

            end

          end
        end
      end

      def find_node_in_mapping(mapping, project_name)
        mapping.each { |k, v|
          if v != nil && v.index(project_name)
            return k
          end
        }
        return nil
      end

      def generate_app_definition_no_mapping(app, directories)
        raise WeaverError.new("Not implemented")
        #        directories.each do |dir|
        #          project_file = File.join(dir, '.project')
        #          classpath_file = File.join(dir, '.classpath')
        #          if File.exists?(project_file) # only process directory if .project file exists
        #            project_xml = File.new(project_file, 'r').read()
        #
        #            # read the name of the project
        #            project_doc = REXML::Document.new(project_xml)
        #            name = REXML::XPath.match(project_doc, 'projectDescription/name/text()')[0].to_s.underscore.symbolize
        #
        #            graph.add_vertex(name)
        #
        #            # find projects that it depends on
        #            project_doc.elements.each('projectDescription/projects/project') do |project|
        #              graph.add_edge(name, project.text().to_s.underscore.symbolize)
        #            end
        #
        #            # find all the above project depends on
        #            if File.exists?(classpath_file)
        #              classpath_xml = File.new(classpath_file, 'r').read()
        #              classpath_doc = REXML::Document.new(classpath_xml)
        #              classpath_doc.elements.each("classpath/classpathentry[@kind='src']") do |e|
        #                path = e.attributes['path'].to_s.underscore
        #                if path[0] == '/'
        #                  graph.add_edge(name, path[1, path.length].symbolize)
        #                end
        #              end
        #            end
        #          end
        #        end
        #        puts graph.write_to_graphic_file
        #        puts graph.topsort_iterator.to_a
      end

      # Finds all the relevant directories to process based on the workingset
      # definition.
      def find_relevant_directories()

      end

      def get_name(project_xml)
        project_doc = REXML::Document.new(project_xml)
        return REXML::XPath.match(project_doc, 'projectDescription/name/text()')[0].to_s
      end

      def create_app_component(app, project_xml, classpath_xml)
        project_doc = REXML::Document.new(project_xml)
        name = REXML::XPath.match(project_doc, 'projectDescription/name/text()')[0].to_s
        puts project_doc
        puts '-----------------'
        dependencies = []
        project_doc.elements.each('projectDescription/projects/project') do |project|
          dependencies << project.text.to_s
        end

        node = Node.new(:node_1, nil)
        component = Component.new(name.underscore.symbolize, nil)
        component.name = name
        component.description = "TODO"
        component.version = "1.0.0"
        node.components[component.id] = component

        app.nodes[node.id] = node
      end

    end

  end

end
