################################################################################
# Licensed Materials - Property of IBM Copyright IBM Corporation 2012. All Rights Reserved.
# U.S. Government Users Restricted Rights - Use, duplication or disclosure restricted by GSA ADP
# Schedule Contract with IBM Corp.
################################################################################

require "logger" # Ruby Logging
require "httpclient"
require "httpclient/http" # Contains a list of HTTP Codes represented as Ruby Constants
require 'json' # JSON parsing
require 'mime/types'
require 'cgi'

require 'devops_deployment_lib/util'

require 'weaver/core/logging'   

################################################################################
# Class Definition                                                             #
################################################################################
module Weaver
  module Icct
  
  # A class that simplifies the interaction with ICCT's REST API.
  #
  # @see http://publib.boulder.ibm.com/infocenter/ieduasst/stgv1r0/index.jsp?topic=/com.ibm.iea.ipas/ipas/1.0/Image_creation_and_customization/IPASv1_ICCT/player.html
  class IcctClient
    include DevopsDeploymentLib::Util
    
    # @private
    # HTTP Header Names
    # =================
    HTTP_ACCEPT = "Accept"
    # @private
    HTTP_CONTENT_TYPE = "Content-Type"
    # @private
    HTTP_ACCEPT_LANGUAGE = "Accept-Language"
    
    # @private
    # HTTP Header Values
    # ==================
    APPLICATION_JSON = "application/json"
    # @private
    TEXT_HTML = "text/html"
    # @private
    ICCT_API_VERSION = "1.2" 
    # @private
    LANGUAGE_EN = "en"

    # @private
    # IWD Resources
    # =============
    ICCT_RESOURCES = "resources"
    # @private
    ICCT_BUNDLES = "#{ICCT_RESOURCES}/bundles"
    # @private
    ICCT_FILES = "#{ICCT_RESOURCES}/files"        
    
    # Class Attributes 
    # ================
    @icct_conf
    @http_client
    @log
    
    ################################################################################
    # Class Methods                                                                #
    ################################################################################
    
      # Create a new instance of the IcctClient.
      #
      # == Params
      #
      # === Required
      # [root_server_uri (URI String)]  The URI String that points to an ICCT server.
      # [user_name (String)]  The user name for the account that will interact with ICCT
      # [password [String]]  The password for the account that will interact with ICCT
      # 
      # === Optional
      # [:log_level (Logger Constant)]  The log level that the object will output at.
      #                                 Valid values are:
      #                                 * Logger::FATAL
      #                                 * Logger::ERROR
      #                                 * Logger::WARN
      #                                 * Logger::INFO
      #                                 * Logger::DEBUG
      # [:log_file (String|STDOUT|STDERR)]  The file to output logs to
      #
      def initialize(root_server_uri=nil, user_name=nil, password=nil, conf={})
        @@logger.debug("Weaver::Icct::IcctClient.new starting");
      
        conf = {
          :log_level => Logger::WARN,
          :log_file => STDOUT
        }.merge(conf)
      
        @log = Logger.new(conf[:log_file])
        @log.level = conf[:log_level]
        
        unless root_server_uri.nil? or user_name.nil? or password.nil?
        
          unless root_server_uri.start_with?("https://")
            # ensure that root_server_uri starts with https://
            root_server_uri = "https://#{root_server_uri}"
          end
        
          @debug_str = ""
          @http_client = HTTPClient.new
          @http_client.debug_dev = @debug_str
          
          @http_client.ssl_config.ssl_version="TLSv1"
          
          @icct_conf = { # default arguments
            :root_server_uri => root_server_uri,
            :user_name => user_name,
            :password => password,
            :header => {
              HTTP_ACCEPT => APPLICATION_JSON,
              HTTP_ACCEPT_LANGUAGE => LANGUAGE_EN
            }
          }
        
          set_icct_user(user_name, password)
        
        else 
          raise ArgumentError, "IcctClient requires values for root_server_uri, user_name, and password"
        end

        @@logger.debug("Weaver::Icct::IcctClient.new ending");
      end
    
      ################################################################################
      # Bundle Interaction Methods                                                   #
      ################################################################################
    
      # Describes all bundles that are available.
      #
      # == Params
      #
      # === Optional
      # [ids (String Array)]  An array of specific bundle IDs to acquire the data for.
      #
      # == Return
      # [Hash Array]  An array of Hash objects representing the content of the bundles.
      #
      def describe_bundles(args={})
        args = {
          :ids => nil
        }.merge(args)
      
        if args[:ids].nil?
          return execute_get_request(ICCT_BUNDLES)
        else
          bundle_arr = []
        
          args[:ids].each do |id|
            bundle = describe_bundle(id)
            bundle_arr.push(bundle)
          end
        
          return bundle_arr
        end
      end # end describe_bundles method
    
      # Describes a single bundle.
      #
      # == Params
      #
      # === Required
      # [id (String)]  The ID of a bundle rce
      #
      # == Return
      # [Hash]  An Hash object representing the content of the bundle
      #
      def describe_bundle(id, args={})
        return execute_get_request("#{ICCT_BUNDLES}/#{id}")
      end # end describe_bundle method
      
      # Create a new bundle.
      #
      # == Params
      # == Required
      # [bundle_attributes (Hash)]  A Hash object containing all of data required to create a new
      #                               bundle.
      #
      # == Return
      # [Hash]  A Hash object containing the data that represents the newly created bundle instance.
      #
      def create_bundle(bundle_attributes, args={})
        execute_post_request(ICCT_BUNDLES, bundle_attributes, args)
      end
      
      # Update a bundle definition
      #
      # == Params
      # == Required
      # [bundle_attributes (Hash)]  A Hash object containing all of data required to create a new
      #                               bundle.
      #
      # == Return
      # [Hash]  A Hash object containing the data that represents the newly created bundle instance.
      #
      def update_bundle(id, bundle_attributes, args={})
         raise WeaverError.new("'id' required to update bundle definition") unless id
         @log.debug("id --> #{id}")
         execute_put_request(ICCT_BUNDLES + "/" + id, bundle_attributes, args)
      end

      def add_files(files)
         boundary = "-------------------------30333176734664"

         body = ""
         files.each do |label, file|
            filename = file.path
            # puts "file has #{file.size} characters"
            content = file.read
            # puts "file contents is #{content.length} characters"
            # Assemble the request body using the special multipart format
            # mime_type = MIME::Types.type_for(filename)[0] || MIME::Types["application/octet-stream"][0]
            mime_type = MIME::Types["application/octet-stream"][0]
            part = "Content-Disposition: form-data; name=\"#{CGI::escape(label)}\"; filename=\"#{ filename }\"\r\n" +
                   "Content-Type: #{ mime_type.simplified }\r\n\r\n#{ content }\r\n"
            body << ("--" + boundary + "\r\n" + part)
         end # end params.each do
         body << ("--" + boundary + "--")
         
         execute_file_post_request(ICCT_FILES, body, {:header => { "Content-Type" => "multipart/form-data; boundary=#{ boundary }"}})
       end # end add_files method

      ################################################################################
      # Helper Methods                                                               #
      ################################################################################
    
      # Set the user account to interact with ICCT.
      #
      # == Params
      #
      # === Required
      # [user_name (String)]  The user name for the account that will interact with ICCT
      # [password [String]]  The password for the account that will interact with ICCT
      #
      def set_icct_user(user_name, password)
         @http_client.set_auth(nil, user_name, password)
         @http_client.www_auth.basic_auth.challenge(@icct_conf[:root_server_uri])
      end # end  set_icct_user method
      
      # Set the SSL verify mode of the HTTP client that will interact with the ICCT system.
      #
      # == Params
      #
      # === Required
      # [verify_mode (OpenSSL Constant)]  The verify mode that should be used to interact 
      #                                   with servers over https. Valid values are:
      #                                   * OpenSSL::SSL::VERIFY_NONE
      #                                   * OpenSSL::SSL::VERIFY_PEER
      #                                   * OpenSSL::SSL::VERIFY_CLIENT_ONCE
      #                                   * OpenSSL::SSL::VERIFY_FAIL_IF_NO_PEER_CERT
      #
      def ssl_verify_mode=(verify_mode)
        @http_client.ssl_config.verify_mode = verify_mode
        @icct_conf[:ssl_verify_mode] = verify_mode
      end # end ssl_verify_mode method

      private
      ################################################################################
      # Private Methods                                                              #
      ################################################################################
    
      # HTTP Request Execution
    
      def execute_get_request(resource, args={})
        args = {
          :header => {}.merge(@icct_conf[:header])
        }.merge(args)
      
        response = execute_request(:get, resource, HTTP::Status::OK, args)
        resp_data = JSON.parse(response.body)
      
        return resp_data
      end # end execute_get_request method
    
      def execute_post_request(resource, body, args={})
        args = {
          :body => body,
          :header => {
            HTTP_CONTENT_TYPE => APPLICATION_JSON
          }.merge(@icct_conf[:header])
        }.merge(args)
      
        response = execute_request(:post, resource, HTTP::Status::CREATED, args)
        resp_data = JSON.parse(response.body)
      
        return resp_data
      end # end execute_post_request method
    
      def execute_file_post_request(resource, body, args={})
        args = {
          :body => body,
          :header => {
            HTTP_CONTENT_TYPE => TEXT_HTML
          }.merge(@icct_conf[:header])
        }.merge(args)
      
        response = execute_request(:post, resource, HTTP::Status::CREATED, args)
        response.body.gsub!(/<.*?>/, "")
        @log.debug("modified response body is #{response.body}")
        resp_data = JSON.parse(response.body)
      
        return resp_data
      end # end execute_post_request method
    
      def execute_put_request(resource, body, args={})
        args = {
          :body => body,
          :header => {
            HTTP_CONTENT_TYPE => APPLICATION_JSON
          }.merge(@icct_conf[:header])
        }.merge(args)
      
        response = execute_request(:put, resource, HTTP::Status::OK, args)
        resp_data = JSON.parse(response.body)
      
        return resp_data
      end # end execute_put_request method
    
      def execute_delete_request(resource, args={})
        args = {
          :header => {}.merge(@icct_conf[:header])
        }.merge(args)
      
        response = execute_request(:delete, resource, HTTP::Status::NO_CONTENT, args)
        resp_data = JSON.parse(response.body)
      
        return resp_data
      end # end execute_delete_request method
    
      def execute_request(http_method, resource, expected_status, args={})
        response = nil
    
        @@logger.debug("execute #{http_method} request on #{@icct_conf[:root_server_uri]}/#{resource}")
        case http_method
          when :get
            response = @http_client.get("#{@icct_conf[:root_server_uri]}/#{resource}", :query => args[:query], :header => args[:header])
          when :post
            response = @http_client.post("#{@icct_conf[:root_server_uri]}/#{resource}", :body => args[:body], :header => args[:header])
          when :delete
            response = @http_client.delete("#{@icct_conf[:root_server_uri]}/#{resource}", :header => args[:header])
          when :put
            response = @http_client.put("#{@icct_conf[:root_server_uri]}/#{resource}", :body => args[:body], :header => args[:header])
        end
      
        @log.debug @debug_str
      
        if response.status == expected_status
          return response
        else
          raise "Error: Unrecognized Code, #{response.code} - #{response.reason}"
        end
      
      end # end execute_request method
      
  end # end IwdVirtualSystemClient class
  end # end Icct module
end # end DevopsDeploymentLib module
