require 'erubis'
require 'string'
require 'json'
require 'weaver/core/logging'   
require 'weaver/core/weaver_error'

module Weaver
  module Icct
    # Generates a ICCT bundle creation JSON request from a Weaver
    # infrastructure definition.
    #
    # Author:: Michael Kalantar
    #
    class IcctRequestGenerator
      # ICCT Operation Types
      # ====================
      IMAGE_BUILD = "imageBuild"
      IMAGE_DEPLOY = "imageDeploy"
      IMAGE_CAPTURE = "imageCapture"
      
      # ICCT Opertation Parameter Passing Styles
      # ========================================
      # cf. StepParameterStyle in StepBean class
      ENVIRONMENT_STYLE = 'environmentStyle'
      SHORT_STYLE = 'shortStyle'
      SHORT_SPACE_STYLE = 'shortSpaceStyle'
      LONG_SPACE_STYLE = 'longSpaceStyle'
      WINDOWS_STYLE = 'windowsStyle'
      
      # ICCT imageDeploy operation extras
      # =================================
      # cf. IBundleConstants w/ prefix ACTIVATION__OPERATIONS
      IMAGE_DEPLOY__OPERATION__TAG = 'tag'
      IMAGE_DEPLOY__OPERATION__SERVICENAME = 'servicename'

      # Creates a new IcctRequestGenerator for a given Environment definition.
      #
      # * *Args*    :
      #   - +name+ -> name of bundle
      #   - +uuid+ -> universal identifier of bundle
      #   - +version+ -> version of bundle
      #   - +description+ -> description of bundle
      #
      # * *Returns* :
      #   - an IcctRequestGenerator instance
      #
      def initialize(bundle_attributes)
         @@logger.debug("Weaver::Icct::IcctRequestGenerator.new starting")
         @@logger.debug(bundle_attributes)
         
         @im_bundle = 'false'
         @products = []
         @requirements = {'os' => [], 'sw' => [], 'bundles' => [], 'hw' => {}}
         @operations = {}

         @id = bundle_attributes['id'] unless bundle_attributes['id'].nil?
         @name = bundle_attributes['name']
         @uuid = bundle_attributes['symbolicName']
         @version = bundle_attributes['version']
         @description = bundle_attributes['version'] 
         @im_bundle = bundle_attributes['isImBundle'] unless bundle_attributes['isImBundle'].nil?
         @products = bundle_attributes['products'] unless bundle_attributes['products'].nil?
         @requirements = bundle_attributes['requirements'] unless bundle_attributes['requirements'].nil?
         @operations = bundle_attributes['operations'] unless bundle_attributes['operations'].nil?

         raise WeaverError.new("Name required to define bundle") unless @name
         raise WeaverError.new("UUID required to define bundle") unless @uuid
         raise WeaverError.new("version required to define bundle") unless @version
                 
         @@logger.debug("Weaver::Icct::IcctRequestGenerator.new ending")
      end # end initialize method
      
      attr_accessor :id

      # Generates a JSON request for a bundle definition.
      #
      # * *Returns* :
      #   - the JSON payload to send to ICCT for creating a bundle
      def to_json()
        @@logger.debug("Weaver::Icct::IcctRequestGenerator.to_json starting") 
        json = {
              'name' => @name,
              'symbolicName' => @uuid,
              'version' => @version,
              'isImBundle' => @im_bundle,
              'products' => @products,
              'requirements' => @requirements,
              'operations' => @operations
               }

        if @description 
           json['description'] = @description
        end # end if @description
        
        if @id
           json['id'] = @id
        end # end if @id

        # puts JSON.pretty_generate(json)
        @@logger.debug("Weaver::Icct::IcctRequestGenerator.to_json ending") 
        return json.to_json()
      end

      def add_product (name, version, publisher = nil)
         raise WeaverError.new("Product name required to define product") unless name
         raise WeaverError.new("Product version required to define product") unless version
         
         product = {'name' => name, 'version' => version}
         product['publisher'] = publisher if publisher
         
         @products << product
      end # end product method
      
      def add_product_requirement (name, version)
         raise WeaverError.new("Product name required to define product requirement") unless name
         raise WeaverError.new("Product version required to define product requirement") unless version
          
         req = {'name' => name, 'minversion' => version }
          
         @requirements['sw'] << req
      end # end add_product_requirement method
      
      def define_role_argument (role_file)
         role_f = role_file.gsub(/\.rb$/, "")
         return {
                        'defaultVal' => role_f,
                        'label' => 'Role file',
                        'isPassword' => false,
                        'type' => 'string',
                        'constraints' => [],
                        'name' => role_f,
                        'direction' => 'input'
                          }
      end # end method
      
      def add_operation (operation_type, rolefile, model, files = [])
         args = []
         args << define_role_argument(rolefile)
         model_role = "ConfigRole"
         model.flatten_children.select { |a| a.kind_of?(Automation) && a.type == :chef_automation }.each do |automation|
            automation.amd.each do |id, attribute|
               # puts "#{attribute.class}:::#{id} --> #{attribute}"
               if id == :role
                  model_role = attribute.value
               end
               if attribute.name.nil? 
                  @@logger.warn("Automation attribute '#{id} has no name; skipping")
               elsif attribute.mutable
                  arg = {
                        'defaultVal' => attribute.value,
                        'label' => attribute.label,
                        'isPassword' => false,
                        'type' => 'string',
                        'constraints' => [],
                        'name' => attribute.name,
                        'direction' => 'input'
                          }
                  arg['type'] = convert_attr_type(attribute.value) if attribute.value
                  arg['type'] = 'password' if arg['isPassword']
                  arg['label'] = arg['name'] if arg['label'].nil?
           
                  args << arg
               end # if ... elif 'true' == attribute[:mutable]
            end # automation.amd.each
         end
      
         command = {
               'parameterStyle' => SHORT_SPACE_STYLE,
               'workdir' => "",
               'args' => args,
               'runas' => 'root',
               'runcmd' => 'TODO.sh'
                   }
                   
         file_details = files.collect do |f|
             executable = false;
             index = /.sh$/ =~ f['name']
             executable = true if index
             command['runcmd'] = f['name'] if executable
             {
                'assetLoc' => f['url'],
                'dest' => "${TEMP}",
                'executable' => executable,
                'source' => f['name']
             }
         end # end files.collect block
                   
         operation = {'files' => file_details, :command => command}
         case operation_type 
         when IMAGE_DEPLOY then
            operation['tag'] = model_role
            operation['servicename'] = model_role
         end 
         @operations[operation_type] = [] if @operations[operation_type].nil?
         @operations[operation_type] << operation
      end # end add_operation method
      
      private
      
      def convert_attr_type(value)
        return "number" if value.kind_of?(Fixnum)
        return "string" if value.kind_of?(String)
        return "boolean" if value.kind_of?(TrueClass) ||  value.kind_of?(FalseClass)
      end

    end # end IcctRequestGenerator class
  end # end Icct module
end # end Weaver module
