require 'weaver/core/base'
require 'weaver/dsl/weaver_parser'
require 'weaver/dsl/requires/product_type_handler'

module Weaver
  # @private
  # Code related to the \Zephyr to \Weaver translator WeaverGenerator.
  module Zephyr
    # Zephyr attributes that we do not create new Weaver::Core:Attributes for
    CORE_ATTRIBUTES = Set.new [ 'name', 'linkType', 'displayName', 'uuid', 'versionString' ]
      
    # Attributes to keep even if nil
    UNREAPED_ATTRIBUTES = Set.new [ :version, :name, :description ]
      
    # WeaverGenerator generates Weaver models from Zephyr Topologies.
    #
    # This version uses ad-hoc logic to deal with Zephyr domains rather than a data-driven
    # model-to-model transformation.
    class WeaverGenerator
      include Weaver::Core

      # The Root of the generated document
      attr_reader :root
      
      # Transform a Zephyr .topology file (with optional automation .topology file) to a Weaver model
      #
      # generator_settings values represent information about the context of the Topology that Weaver wishes
      #  to store in the Weaver model but Zephyr didn't store in the Zephyr model.  The keys are
      # :bundle true if the topology represents a bundle (e.g., was stored in a RAM Bundle Asset)
      # :bundle_name the name of the bundle associated with the topology
      # :cloud_type a Symbol for the type of cloud the topology was stored with, e.g. :iwd_vsp
      # :cloud_url the URL of the cloud API
      # :cloud_image_id the ID of the image associated with the topology
      # :virtual_image_name the name of the image associated with the topology
      # :virtual_image_descr the description of the image associated with the topology
      # :cloud_image_osversion the OS version if the cloud metadata provided it
      # :cloud_metals the 'metals', e.g. 'BRONZE/1/2', 'SILVER/3/4' that the image is offered in
      #
      def self.transform(generator_settings, zephyr_file, zephyr_automation_file = nil)
        raise "generator settings should be a Hash" unless generator_settings.kind_of?(Hash)
        if zephyr_automation_file.nil?
          @@logger.debug "Transforming #{zephyr_file.path}"
        elsif
        @@logger.debug "Transforming #{zephyr_file.path} with #{zephyr_automation_file.path}"
        end

        topology_xml = File.new(zephyr_file.path, 'r').read()
        doc = REXML::Document.new(topology_xml)
        generator = WeaverGenerator.new(doc, zephyr_file, generator_settings)

        if zephyr_automation_file.nil?
          generator.transform_document(nil)
        else
          automation_topology_xml = File.new(zephyr_automation_file.path, 'r').read()
          automation_doc = REXML::Document.new(automation_topology_xml)
          generator.transform_document(automation_doc)
        end

        return generator.root
      end

      # Users should call WeaverGenerator.transform() instead of constructing instances
      def initialize(doc, root_id, generator_settings)
        raise "generator settings should be a Hash" unless generator_settings.kind_of?(Hash)

        @zephyr_doc = doc
        @entities = {}
        @conceptuals = {}
        @hosting_links = []
        @member_links = []
        @constraint_links = []
        @realization_links = []
        @installed_units = Set.new
        @to_be_installed_units = Set.new
        @parameters = {}
        @generator_settings = generator_settings
        @operation_groups = []
        @realizations = {}

        @os_handler = Weaver::Core::Os::OperatingSystemHandler.new
        
        # Load the standard Weaver types into 'root_entity'
        @root_entity = Weaver::Dsl::WeaverParser.new().parse(nil)
        @root = @root_entity.clone
        @root.id = root_id 
      end
      
      def transform_document(automation_document)

        # Note that we aren't checking namespaces
        raise "Not a topology" unless @zephyr_doc.root().name == "topology"

        @@logger.debug "Transforming semantic topology"
        transform_topology(@zephyr_doc.root())
        if !automation_document.nil?
          @@logger.debug "Transforming automation topology"
          transform_topology(automation_document.root())
        end
        link_entities()

        # At this point
        #   - if we have only 'installed' units we are an infrastructure
        #   - if we have only 'conceptual' we are an application, or if generator_settings[:bundle]
        #   - if we have both and not generator_settings[:bundle] we are an environment
        #   - if we have neither we are an empty environment
        @@logger.debug "#{@conceptuals.size} conceptual units, #{@installed_units.size} installed units, #{@to_be_installed_units.size} to-be-installed units"
        @@logger.debug "The conceptuals are: #{@conceptuals.keys.inspect}"
        if @installed_units.size > 0 && @conceptuals.size == 0
          infrastructure = Topology.new(@zephyr_doc.root().attributes['name'].symbolize, @root)
          infrastructure.type = :infrastructure
          @root.topologies[infrastructure.id] = infrastructure
          infrastructure.name = @zephyr_doc.root().attributes['displayName']
          infrastructure.description = @zephyr_doc.root().attributes['description'] || 'generated'
        elsif (@installed_units.size == 0 && @conceptuals.size > 0) || @generator_settings[:bundle]
          application = Topology.new(@zephyr_doc.root().attributes['name'].symbolize, @root)
          application.type = :application
          application.name = @zephyr_doc.root().attributes['displayName']
          application.description = @zephyr_doc.root().attributes['description']
          @root.topologies[application.id] = application 
        else
          env = Topology.new(@zephyr_doc.root().attributes['name'].symbolize, @root)
          env.type = :environment
          env.name = @zephyr_doc.root().attributes['displayName']
          env.description = @zephyr_doc.root().attributes['description']
          env.topologies[:infrastructure] = Topology.new(:infrastructure, @root.env)
          env.topologies[:infrastructure].type = :infrastructure
          @root.topologies[env.id] = env
          infrastructure = env.infrastructure

          if @conceptuals.size > 0
            application = Topology.new(:application, @root.env)
            env.topologies[:application] = application
          end
        end

        if !infrastructure.nil?
          cloud = Cloud.new(@generator_settings[:cloud_type])
          cloud.url = @generator_settings[:cloud_url]
          infrastructure.cloud = cloud
        end

        if @virtual_system.nil?
          # No VirtualSystemUnit, treat as physical topology with each Zephyr OS becoming a Node
          add_oses_as_nodes(infrastructure, application)
          @@logger.debug "No virtual system, adding OSes as nodes in a distributed application"

          if @parameters.size > 0
            @@logger.warn "Ignoring #{parameters.size} parameters in non-cloud topology"
          end
        else
          @@logger.debug "Adding OSes to virtual system"
          
          # VirtualSystemUnit exists, it becomes the Node, the other OSes become the OS
          add_operation_groups_to_virtual_system()
          add_os_to_virtual_system(infrastructure.nil? ? application : infrastructure)

          # Add deploy-time parameters to infrastructure
          @virtual_system.amd.merge!(@parameters)
          @parameters.each { |id, parameter|
            parameter.scope = @virtual_system
#            @virtual_system.unresolved_attributes[id] = parameter.default
          }
          
          cloud_merge()
        end

        @@logger.debug "Looking for ignored units to document without modeling them"
        add_ignored_units(infrastructure, application)

      end

      protected
      
      # :stopdoc:
      def cloud_merge
        cloud = Cloud.new(@generator_settings[:cloud_type])
        cloud.image = @generator_settings[:cloud_image_id]
        @virtual_system.cloud = cloud
        @virtual_system.name = @generator_settings[:virtual_image_name] unless @generator_settings[:virtual_image_name].nil?
        @virtual_system.name = @generator_settings[:bundle_name] unless @generator_settings[:bundle_name].nil?
        @virtual_system.description = @generator_settings[:virtual_image_descr] unless @generator_settings[:virtual_image_descr].nil?

        # Override topology OS version using cloud metadata
        @virtual_system.os.version = @generator_settings[:cloud_image_osversion] unless @generator_settings[:cloud_image_osversion].nil?

        if @generator_settings[:cloud_metals]
          attr_metal = Attribute.new(:metal, @virtual_system)
          attr_metal.label = 'Instance Metal'
          attr_metal.default = @generator_settings[:cloud_metals][0]
          attr_metal.valid_values = @generator_settings[:cloud_metals]
          attr_metal.mutable = true
          @virtual_system.amd[:metal] = attr_metal
          @virtual_system.metal = attr_metal.default
        end

#        attr_ip = Attribute.new(:ip_address, @virtual_system)
#        attr_ip.label = 'IP Address'
#        attr_ip.mutable = true
#        @virtual_system.amd[:ip_address] = attr_ip
#        @virtual_system.ip_address = nil
      end

      def add_ignored_units(infrastructure, application)
        # Add comments for unmodeled types
        @entities.select { |id, weaver_entity| weaver_entity.kind_of?(IgnoredUnit) || weaver_entity.kind_of?(UnmodeledConcept) }.each do |id, weaver_entity|
          container = part_of_infrastructure(weaver_entity) ? infrastructure : application
          container = container || infrastructure || application  # If we guessed a nil container pick another
          container.inner_comments << weaver_entity.note if weaver_entity.kind_of?(UnmodeledConcept)
#          unless weaver_entity.description.nil?
#            weaver_entity.description.each_line { |line| container.inner_comments << line.chomp } 
#          end
        end
        
        # Add groups
        @entities.select { |id, weaver_entity| weaver_entity.kind_of?(Group) }.each do |id, weaver_entity|
          container = part_of_infrastructure(weaver_entity) ? infrastructure : application
          container = container || infrastructure || application  # If we guessed a nil container pick another
          if container.is_a?(Weaver::Core::GroupContainerMixin)
            container.groups[id] = weaver_entity
          else
            @@logger.warn "Can't create group #{weaver_entity.id} in container type #{container.class}"
          end
        end
        
        # Add users
        @entities.select { |id, weaver_entity| weaver_entity.type == :user }.each do |id, weaver_entity|
          # Currently we only handle users on Application, not Infrastructure
          if application.nil?
            infrastructure.users[id] = weaver_entity
          else
            application.users[id] = weaver_entity
          end
        end
      end
      
      def add_oses_as_nodes(infrastructure, application)
        @entities.each do |id, weaver_entity|
          if weaver_entity.kind_of?(Node)
            if part_of_infrastructure(weaver_entity)
              # TODO Remove
              if infrastructure.nil?
                raise "Can't store #{weaver_entity} in nil infrastructure!"
              end

              weaver_entity.cloud = Cloud.new(@generator_settings[:cloud_type])
              infrastructure.nodes[id] = weaver_entity
              weaver_entity.scope = infrastructure
            else
              application.nodes[id] = weaver_entity
              weaver_entity.scope = application
            end
          else
            if weaver_entity.kind_of?(Component) && weaver_entity.scope.nil?
              @@logger.debug "Ignoring unhosted Component #{weaver_entity}" unless weaver_entity.type == :group
            end
          end
        end
      end

      def add_operation_groups_to_virtual_system
        @operation_groups.each do |og|
          if og.ports.size > 0
            if @virtual_system.firewall.nil?
              @virtual_system.firewall = Firewall.new()
            end
            @virtual_system.firewall.tcp.push(FirewallRule.new(og.ports))
          end
          
          # We currently only generate script metadata from <RunOperation>s for bundle topologies
          if @generator_settings[:bundle] && og.scripts.size > 0
            @@logger.debug "Found #{og.scripts.size} scripts on Operation Group"
            bundle_component = @virtual_system.components[:bundle_component]
            if bundle_component.nil?
              bundle_component = Component.new(:bundle_component, og)
              @virtual_system.components[bundle_component.id] = bundle_component
            end
            cnt = 1
            og.scripts.each do |run_operation|
              script = Automation.new(('script' + cnt.to_s).symbolize, bundle_component)
              script.executable = run_operation.run_commands
              script.parameter_style = run_operation.parameter_style
              script.type = run_operation.interpreter
              script.run_as = run_operation.run_as
              # Script does not model Zephyr run_at nor working_directory
              run_operation.artifacts.each { |artifact|
                script.inner_comments.push("Deploy '#{artifact[:uri]}' on '#{artifact[:deployLocation]}'#{artifact[:executable_flag] ? ' and set to executable' : ''}")
              }
              
              bundle_component.automation[script.id] = script
              cnt = cnt+1
            end
          end
        end

      end
      
      # vs_container will be Infrastructure or Application
      def add_os_to_virtual_system(vs_container)
        @entities.each { |id, weaver_entity|
          if weaver_entity.kind_of?(Node)
            if weaver_entity.equal?(@virtual_system)
              vs_container.nodes[id] = weaver_entity
              weaver_entity.scope = vs_container 
            else
              merge_osnode_with_virtual_system(weaver_entity)
            end
          else
            if weaver_entity.kind_of?(Component) && weaver_entity.scope.nil?
              @@logger.debug "Ignoring unhosted Component #{weaver_entity}" unless weaver_entity.type == :group
            end
          end
        }
      end

      def merge_osnode_with_virtual_system(osnode)
        # merge components
        osnode.components.each { |id, component|
          @virtual_system.components[id] = component
        }

        # merge OS if this is the most-concrete Zephyr OS
        if osnode.realized_by.empty?
          @virtual_system.os = osnode.os
        end
      end

      def part_of_infrastructure(weaver_entity)
        return !@conceptuals.key?(weaver_entity.id) || @installed_units.member?(weaver_entity.id)
      end

      def transform_topology(topology_element)
        topology_element.elements.each { |ele|
          case ele.name
          when 'unit.softwareInstallUnit'
            handle_software_install_unit(ele)
          when 'unit.linuxLocalUser'
            handle_linux_local_user_unit(ele)
          when 'unit.user'
            handle_user_unit(ele)
          when 'unit.operatingSystemUnit'
            handle_generic_operating_system_unit(ele)
          when 'unit.linuxOperatingSystemUnit'
            handle_linux_operating_system_unit(ele)
          when 'unit.zOS'
            handle_z_operating_system_unit(ele)
          when 'unit.db2SystemUnit'
            handle_db2_system_unit(ele)
          when 'unit.x86ServerUnit'
            handle_x86_server_unit(ele)
          when 'unit.powerServerUnit'
            handle_power_server_unit(ele)
          when 'unit.zCP'
            handle_z_cp_unit(ele)
          when 'unit.lpar'
            handle_lpar_unit(ele)
          when 'unit.zServer'
            handle_z_server_unit(ele)
          when 'unit.aixLocalUser'
            handle_aix_local_user_unit(ele)
          when 'unit.aixOperatingSystemUnit'
            handle_aix_operating_system_unit(ele)
          when 'contract.explicit'
            handle_contract_explicit(ele)
          when 'unit.windowsLocalGroup'
            handle_unit(ele)              # TODO implement
          when 'unit.windowsOperatingSystemUnit'
            handle_unit(ele)              # TODO implement
          when 'unit.windowsLocalUser'
            handle_unit(ele)              # TODO implement
          when 'unit.zLinux'
            handle_unit(ele)              # TODO implement
          when 'unit.ihsSystemUnit'
            handle_ihs_system_unit(ele)
          when 'unit.wasSystemUnit'
            handle_was_system_unit(ele)
          when 'unit.wasAppServer'
            handle_was_app_server_unit(ele)
          when 'unit.wasClusterUnit'
            handle_unit(ele)              # TODO implement
          when 'unit.wasNodeUnit'
            handle_was_node_unit(ele)
          when 'unit.portConfigUnit'
            handle_port_config_unit(ele)
          when 'unit.wasClassLoaderConfigurationUnit'
            handle_unit(ele)              # TODO implement
          when 'unit.wasCellUnit'
            handle_was_cell_unit(ele)
          when 'unit.wasNodeGroupUnit'
            handle_unit(ele)              # TODO implement
          when 'unit.umlActorUnit'
            handle_unit(ele)              # TODO implement
          when 'unit.httpServerUnit'
            handle_unit(ele)              # TODO implement
          when 'unit.databaseInstanceUnit'
            handle_unit(ele)              # TODO implement
          when 'unit.databaseUnit'
            handle_database_unit(ele)
          when 'unit.databaseSystemUnit'
            handle_database_system_unit(ele)
          when 'link.realization'
            handle_link_realization(ele)
          when 'unit.serverUnit'
            handle_server_unit(ele)
          when 'unit.db2InstanceUnit'
            handle_db2_instance_unit(ele)
          when 'unit.userGroup'
            handle_unit(ele)              # TODO implement
          when 'unit'
            handle_generic_unit(ele)      # TODO implement
          when 'unit.runOperation'
            handle_run_operation_unit(ele)
          when 'unit.virtualSystem'
            handle_virtual_system_unit(ele)
          when 'unit.operationGroup'
            handle_operation_group_unit(ele)
          when 'unit.location'
            handle_location_unit(ele)
          when /^unit\./
            @@logger.debug "Default processing for Unit <#{ele.name}>"
            handle_unit(ele)              # TODO implement
          when 'annotation'
            handle_annotation(ele, @root)
          when /^constraint\./
            handle_constraint_other(ele, @root)
          when 'component.jar'
            handle_unit(ele)
          when 'ldapServerUnit'
            handle_unit(ele)
          when 'softwareComponent'
            handle_software_install_unit(ele)
          when /^component\./
            handle_unit(ele)
          when 'import'
            handle_import(ele)
          when 'attributeMetaData'
            handle_amd(ele, @root)
          when 'extendedAttribute'
            handle_extended_attribute(ele, @root)
          when /^artifact\./
            @@logger.debug "Ignoring artifact <#{ele.name}>"
          when 'link.constraint'
            handle_link_constraint(ele, @root)
          when 'link.member'
            handle_link_member(ele, @root)
          when 'link.hosting'
            handle_link_hosting(ele, @root)
          when 'link.dependency'
            handle_link_dependency(ele, @root)
          else
            raise "unknown element #{ele.name}"
          end
        }

        return @root
      end

      def handle_generic_unit(unit)
        # <core:unit goalInstallState='installed' initInstallState='installed' name='bundle_110487'>
        handle_unit(unit)
      end

      def handle_location_unit(unit)
        unmodeled_concept = entity_from_unit(unit, UnmodeledConcept)
        unmodeled_concept.note = "location :#{unmodeled_concept.id} { ... }"
        unit.elements.each { |ele| handle_unit_subtag(ele, unit, unmodeled_concept) }
      end
      
      # Catch-all handler; 'unit' is the Zephyr Unit's XML Element
      def handle_unit(unit)
        dummy = entity_from_unit(unit, IgnoredUnit)
        unit.elements.each { |ele| handle_unit_subtag(ele, unit, dummy) }
      end

      # ele is the XML Element
      # unit is the Zephyr Unit's XML Element
      # weaver_entity is the Weaver Entity represented by 'unit'
      def handle_unit_subtag(ele, unit, weaver_entity)
        case ele.name
        when /capability\./
          handle_capability(ele, weaver_entity)
        when 'capability'
          handle_capability(ele, weaver_entity)
        when 'requirement'
          handle_requirement(ele, weaver_entity)
        when 'link.hosting'
          handle_link_hosting(ele, weaver_entity)
        when 'link.member'
          handle_link_member(ele, weaver_entity)
        when 'link.realization'
          handle_link_realization(ele, weaver_entity)
        when 'artifact.imageFile'
          handle_file_artifact(ele, weaver_entity)
        when /^artifact\./
          @@logger.debug "Ignoring artifact <#{ele.name}>"
        when 'link.constraint'
          handle_link_constraint(ele, weaver_entity)
        when 'annotation'
          handle_annotation(ele, weaver_entity)
        when 'constraint.attributePropagation'
          handle_attribute_propagation(ele, weaver_entity)
        when 'constraint.capacity'
          handle_constraint_capacity(ele, nil, weaver_entity)
        when 'reference'
          handle_reference(ele, weaver_entity)
        when /^service\./
          handle_service(ele, weaver_entity)
        when 'service'
          handle_service(ele, weaver_entity)
        when /^constraint\./
          handle_constraint_other(ele, weaver_entity)
        when /^analysisConstraint\./
          handle_constraint_other(ele, weaver_entity)
        when 'stereotype'
          handle_stereotype(ele, weaver_entity)
        else
          # TODO have a strict/nonstrict flag and ignore when in nonstrict mode
          raise "unknown element #{ele.name} on #{unit.name}: #{ele}"
        end
      end

      def handle_software_install_unit(siu)
        component = entity_from_unit(siu, :software_product)
        
        siu.elements.each do |ele|
          case ele.name
          when 'capability.softwareInstall'
            handle_software_install(ele, component)
          when 'capability.version'
            handle_version(ele, component)
          else
            handle_unit_subtag(ele, siu, component)
          end
        end
        
        # Remove any attributes that don't have validators or values, assuming they are useless/meaningless (e.g. :majorversion)
        component.amd.reject! do |id, attr| 
          attr.unresolved_value.nil? &&     # Don't reject if there is a value or a late_binding 
            attr.validators.length == 0 &&  # Don't reject if there is a validator
            attr.default.nil? &&            # Don't reject if there is a default
            attr.name.nil? &&               # Don't reject if there is a link to a deployment parameter deployment name
            !UNREAPED_ATTRIBUTES.member?(attr.id)
        end
      end

      def handle_user_unit(uu)
        # TODO We can't store this User until we refactor to allow Infrastructure Users

        user = entity_from_unit(uu, User)

        uu.elements.each { |ele|
          case ele.name
          when 'capability.user'
            handle_local_user(ele, user)
          else
            handle_unit_subtag(ele, uu, user)
          end
        }
      end

      def handle_aix_local_user_unit(aluu)
        # TODO We can't store this User until we refactor to allow Infrastructure Users

        user = entity_from_unit(aluu, User)

        aluu.elements.each { |ele|
          case ele.name
          when 'capability.aixLocalUser'
            handle_local_user(ele, user)
          when 'requirement'
            handle_requirement(ele, user)
          else
            raise "unknown User child element #{ele.name} on #{aluu.name}"
          end
        }
      end

      def handle_linux_local_user_unit(lluu)
        # TODO We can't store this User until we refactor to allow Infrastructure Users

        user = entity_from_unit(lluu, User)

        lluu.elements.each { |ele|
          case ele.name
          when 'capability.suseLinuxLocalUser'
            handle_local_user(ele, user)
          when 'capability.redhatLinuxLocalUser'
            handle_redhat_user(ele, user)
          when 'capability.linuxLocalUser'
            handle_local_user(ele, user)
          else
            handle_unit_subtag(ele, lluu, user)
          end
        }
      end

      def handle_database_system_unit(dsu)
        # <database:unit.databaseSystemUnit conceptual='true' displayName='Database Installation'
        #   name='DatabaseSystemUnit_517802'>
        # TODO Create a sub-type of Component for Database Component

        component = entity_from_unit(dsu, Component)
        dsu.elements.each do |ele|
          case ele.name
          when 'capability.databaseSystem'
            handle_capability(ele, component) # TODO domain-specific handling
          else
            handle_unit_subtag(ele, dsu, component)
          end
        end
      end

      def handle_port_config_unit(pcu)
        # <os:unit.portConfigUnit configurationUnit='true' displayName='Port Configuration' name='PortConfigUnit_489854' uuid='_i7FQ8HHSEeG1mfyCaLD8og'> (RuntimeError)
        #  <os:capability.port applicationLayerProtocol='HTTP' displayName='dcs_unicast_address' linkType='dependency' name='dcs_unicast_address' port='9353' portName='dcs_unicast_address' transportLayerProtocol='TCP' uuid='_i7FQ8XHSEeG1mfyCaLD8og'>
        #   <core:attributeMetaData attributeName='portName' mutable='true' uuid='_i7FQ8nHSEeG1mfyCaLD8og'/>
        #  </os:capability.port>
        # </os:unit.portConfigUnit>
        
        ports = entity_from_unit(pcu, Component)
        ports.description = ports.description || 'TCP Port usage'
        ports.amd[:port] = Attribute.new(:port, ports)
        ports.amd[:port].value = []
        
        pcu.elements.each do |ele|
          case ele.name
          when 'capability.port'
            if ele.attributes['transportLayerProtocol'] == 'TCP'
              ports.amd[:port].value.push(ele.attributes['port'].to_i)
            else
              @@logger.warn "Unexpected transportLayerProtocol #{ele.attributes['transportLayerProtocol']}"
            end
          else
            handle_unit_subtag(ele, pcu, ports)
          end
        end
      end
      
      def handle_was_node_unit(wnu)
        # <was:unit.wasNodeUnit name='WasNodeUnit_489854'>
        #  <was:capability.wasNode displayName='WebSphereUnmanagedNode' isDefaultProfile='true' linkType='any' name='WebSphereNode' nodeName='node1' profileLocation='/Program Files\IBM\WebSphere\AppServer\profiles\AppSrv01' profileName='AppSrv01' profileType='default' wasVersion='6.1'/>
        #  <was:capability.wasSubstitutionVariable displayName='Substitution variable' linkType='any' name='Substitution Variable'/>
        #  <was:capability.wasConfigurationContainer displayName='Configuration Container' linkType='any' name='Provide Configuration Container'/>
        #  <was:capability.wasSharedLibContainer displayName='Provide shared library Container' linkType='any' name='SharedlibraryContainer'/>
        #  ..
        # </was:unit.wasNodeUnit>

        wasnode = entity_from_unit(wnu, :was_node_component)
        
        wnu.elements.each do |ele|
          case ele.name
          when 'capability.wasNode'
            handle_was_node(ele, wasnode) # TODO domain-specific handling
          else
            handle_unit_subtag(ele, wnu, wasnode)
          end
        end
      end
      
      def handle_was_node(capability, wasnode)
        # <was:capability.wasNode displayName='WebSphereUnmanagedNode' isDefaultProfile='true' linkType='any' name='WebSphereNode' nodeName='node1' profileLocation='/Program Files\IBM\WebSphere\AppServer\profiles\AppSrv01' profileName='AppSrv01' profileType='default' wasVersion='6.1'/>

        wasnode.is_default_profile = capability.attributes['isDefaultProfile'] == 'true'
        wasnode.node_name =        capability.attributes['nodeName']
        wasnode.profile_location = capability.attributes['profileLocation']
        wasnode.profile_name =     capability.attributes['profileName']
        wasnode.profile_type =     capability.attributes['profileType'].symbolize unless capability.attributes['profileType'].nil? 
        
        handle_capability(capability, wasnode, Set.new([ 'isDefaultProfile', 'nodeName', 'profileLocation', 'profileName', 'profileType', 'wasVersion' ]))
      end
      
      def handle_was_cell_unit(wcu)
        # <was:unit.wasCellUnit configurationUnit='true'>
        #  <was:capability.wasCell cellName='cell1' displayName='WebSphere 6.1 Stand-Alone Cell' isDistributed='false' linkType='any' name='WebSphere61StandAloneCell' wasVersion='6.1'/>
        #  <was:capability.wasSubstitutionVariable displayName='Substitution variable' linkType='any' name='Substitution Variable'/>
        #  <was:capability.wasConfigurationContainer displayName='WebSphere Cell Configuration Container' linkType='any' name='WebSphereCellConfigurationContainer'/>
        #  <was:capability.wasSharedLibContainer displayName='WebSphere Cell Shared Library Container' linkType='any' name='WebSphereCellSharedLibraryContainer'/>
        #   <core:requirement displayName='Nodes in the Cell' dmoType='was:WasNodeUnit' linkType='member' mutable='true' name='NodeMember'>
        #    <constraint:constraint.memberCardinality maxValue='2' minValue='1' .../>
        #  </core:requirement>
        #  <core:requirement displayName='Node Groups in the Cell' dmoType='was:WasNodeGroupUnit' linkType='member' ...>
        #   <constraint:constraint.memberCardinality maxValue='1' minValue='1' name='memberCardinality'/>
        #  </core:requirement>
        #  ...
        #  <core:link.member source='#/WasCellUnit_263533' target='#/WasNodeGroupUnit_263533'/>
        #  <core:link.member source='#/WasCellUnit_263533' target='#/WasNodeUnit_263533'/>
        # </was:unit.wasCellUnit>
        
        group = entity_from_unit(wcu, :was_cell)
        
        wcu.elements.each do |ele|
          case ele.name
          when 'capability.wasCell'
            handle_was_cell(ele, group)
          else
            handle_unit_subtag(ele, wcu, group)
          end
        end
      end
      
      def handle_was_cell(capability, wascell)
        # <was:capability.wasCell cellName='cell1' displayName='WebSphere 6.1 Stand-Alone Cell' isDistributed='false' linkType='any' name='WebSphere61StandAloneCell' wasVersion='6.1'/>

        wascell.cell_name =       capability.attributes['cellName']
        
        handle_capability(capability, wascell, Set.new([ 'cellName', 'wasVersion' ]))
      end
      
      def handle_was_app_server_unit(wasu)
        # <was:unit.wasAppServer name='WebsphereAppServerUnit_263533'>

        # TODO Create a sub-type of Component for WebSphere App Server Component

        component = entity_from_unit(wasu, Component)

        wasu.elements.each do |ele|
          case ele.name
          when 'capability.j2eeContainer'
            handle_capability(ele, component) # TODO domain-specific handling
          when 'capability.servletContainer'
            handle_capability(ele, component) # TODO domain-specific handling
          when 'capability.jspContainer'
            handle_capability(ele, component) # TODO domain-specific handling
          when 'capability.ejbContainer'
            handle_capability(ele, component) # TODO domain-specific handling
          when 'capability.jcaContainer'
            handle_capability(ele, component) # TODO domain-specific handling
          when 'capability.wasServer'
            handle_capability(ele, component) # TODO domain-specific handling
          when 'capability.wasSubstitutionVariable'
            handle_capability(ele, component) # TODO domain-specific handling
          when 'capability.wasConfigurationContainer'
            handle_capability(ele, component) # TODO domain-specific handling
          when 'capability.wasSharedLibContainer'
            handle_capability(ele, component) # TODO domain-specific handling
          when 'capability.portConsumer'
            handle_capability(ele, component) # TODO domain-specific handling
          when 'capability.jre'
            handle_capability(ele, component) # TODO domain-specific handling
          when 'capability.wasModuleClassLoaderPolicy'
            handle_capability(ele, component) # TODO domain-specific handling
          when 'capability.wasApplicationClassLoaderPolicy'
            handle_capability(ele, component) # TODO domain-specific handling
          when 'capability.wasSecurity'
            handle_capability(ele, component) # TODO domain-specific handling
          else
            handle_unit_subtag(ele, wasu, component)
          end
        end
      end

      def handle_ihs_system_unit(isu)
        # <ihs:unit.ihsSystemUnit conceptual='true' displayName='IBM HTTP Server 6.1' name='IhsSystemUnit_5389'>
        #  <ihs:capability.ihsSystem displayName='IHS System' linkType='any' name='IHS System'/>
        #  ...
        # </ihs:unit.ihsSystemUnit>
        
        component = entity_from_unit(isu, Component)
        
        isu.elements.each do |ele|
          case ele.name
          when 'capability.ihsSystem'
            handle_unit_subtag(ele, isu, component) # add extra IHS-specific logic?
          else
            handle_unit_subtag(ele, isu, component)
          end
        end
      end
      
      def handle_was_system_unit(wsu)
        # <was:unit.wasSystemUnit name='WasSystemUnit_209789'>

        component = entity_from_unit(wsu, :was_system_component)

        wsu.elements.each { |ele|
          case ele.name
          when 'capability.wasSystem'
            handle_was_system(ele, component)
          when 'capability.version'
            handle_version(ele, component)
          else
            handle_unit_subtag(ele, wsu, component)
          end
        }
      end

      def handle_database_unit(du)
        # <database:unit.databaseUnit conceptual='true' displayName='Database' name='DatabaseUnit_517802'>
        #  <database:capability.database linkType='any' name='Database'/>
        #   ...
        # </database:unit.databaseUnit>
        
        # TODO Create a sub-type of Component for Database System Component
        component = entity_from_unit(du, Component)
        du.elements.each { |ele| handle_unit_subtag(ele, du, component) }
      end
      
      def handle_db2_instance_unit(diu)
        # <db2Deploy:unit.db2InstanceUnit name='DB2InstanceUnit_530096' uuid='_o63CEHHSEeG1mfyCaLD8og'> (RuntimeError)
        #  <db2Deploy:capability.db2Instance autoStart='false' db2InstanceName='DB2INST' db2Version='9.5' displayName='DB2 Instance' linkType='any' name='DB2 Instance' port='50000' uuid='_o63CEXHSEeG1mfyCaLD8og'>
        #   <core:attributeMetaData attributeName='password' encrypted='true' uuid='_o63CEnHSEeG1mfyCaLD8og'/>
        #  </db2Deploy:capability.db2Instance>
        #  ...
        # </db2Deploy:unit.db2InstanceUnit>
        
        # TODO Create a sub-type of Component for DB2 System Component
        component = entity_from_unit(diu, Component)
        component.name = 'DB2'
        component.description = component.description || 'DB2 Instance'

        diu.elements.each do |ele|
          case ele.name
          when 'db2Deploy:capability.db2Instance'
            handle_unit_subtag(ele, diu, component)      # TODO improve?
          else
            handle_unit_subtag(ele, diu, component)
          end
        end
      end
      
      def handle_db2_system_unit(dsu)
        # <db2Deploy:unit.db2SystemUnit displayName='DB2 System' name='DB2SystemUnit_523779'>

        # TODO Create a sub-type of Component for DB2 System Component
        component = entity_from_unit(dsu, :software_product)
        component.description = component.description || 'DB2 System'

        dsu.elements.each do |ele|
          case ele.name
          when 'capability.db2JdbcDriver'
            handle_db2_jdbc_driver(ele, component)
          when 'capability.db2System'
            handle_db2_system(ele, component)
          else
            handle_unit_subtag(ele, dsu, component)
          end
        end
      end

      def handle_x86_server_unit(xsu)
        # <server:unit.x86ServerUnit displayName='x86 Server' name='X86ServerUnit_992066'>

        # TODO Create a Weaver entity (Node?) to hold capability values
        node = entity_from_unit(xsu, Server)

        xsu.elements.each { |ele|
          case ele.name
          when 'capability.x86Server'
            handle_x86_server(ele, node)
          else
            handle_unit_subtag(ele, xsu, node)
          end
        }

      end

      def handle_operation_group_unit(ogu)
        # <appliance:unit.operationGroup name='operationGroupU'>
        #  <appliance:capability.operationGroup groupName='IconTask' name='group' source='com.ibm.icon.abstraction.rc2.hslt_1.0.0.0'>
        #   <core:annotation context='software_install' uuid='_3ZU811XxEeGKQaS3NjuyCQ'>
        #    <core:detail key='dmo_path' uuid='_3ZU82FXxEeGKQaS3NjuyCQ' value='/SoftwareInstallUnit_872768'/>
        #   </core:annotation>
        #   ...
        #  </appliance:capability.operationGroup>
        # </appliance:unit.operationGroup name='operationGroupU'>

        og = entity_from_unit(ogu, OperationGroup)
        @operation_groups.push(og)
        
        ogu.elements.each do |ele|
          case ele.name
          when 'capability.operationGroup'
            handle_capability(ele, og)      # TODO Extract more things from the OperationGroup
          when 'capability.portForward'
            handle_port_forward(ele, og)
          else
            handle_unit_subtag(ele, ogu, og)
          end
        end
      end
      
      def handle_virtual_system_unit(vsu)
        # <appliance:unit.virtualSystem uuid="_ow3HUk6NEeG0Fozkp1X9yQ" name="VirtualSystemUnit_554846">
        #   <appliance:capability.virtualSystem uuid="_ow3HU06NEeG0Fozkp1X9yQ" name="virtualSystem"/>
        #   ...
        # </appliance:unit.virtualSystem>

        vs = entity_from_unit(vsu, Node)
        if @virtual_system.nil?
          @virtual_system = vs
        else
          @@logger.warn "Multiple Virtual Systems in topology"
        end

        vsu.elements.each do |ele|
          case ele.name
          when 'capability.virtualSystem'
            handle_capability(ele, vs)  # TODO model-specific handler needed
          else
            handle_unit_subtag(ele, vsu, vs)
          end
        end
      end

      def handle_run_operation_unit(rou)
        # <appliance:unit.runOperation name="runOperationUni" goalInstallState="installed" initInstallState="installed">
        #   <appliance:capability.runOperation name="operation" operationName="New Step"
        #      generated="false" parameterStyle="environment" runAt="imageBuild"
        #      interpreter="/bin/sh" runAsUser="root" runCommands="dummy.sh" workingDirectory="/tmp"/>
        # </appliance:unit.runOperation>

        op = entity_from_unit(rou, RunOperation)

        rou.elements.each { |ele|
          case ele.name
          when 'capability.runOperation'
            handle_run_operation(ele, op)
          else
            handle_unit_subtag(ele, rou, op)
          end
        }
      end

      def handle_server_unit(su)
        # <server:unit.serverUnit conceptual='true' goalInstallState='unknown' initInstallState='unknown' name='serverUnit'>

        node = entity_from_unit(su, Server)

        su.elements.each { |ele|
          case ele.name
          when 'capability.server'
            handle_server(ele, node)
          else
            handle_unit_subtag(ele, su, node)
          end
        }
      end

      def handle_z_server_unit(zsu)
        # <systemz:unit.zServer conceptual='true' displayName='System z Server' name='ZServerUnit_8549'>

        node = entity_from_unit(zsu, Server)

        zsu.elements.each { |ele|
          case ele.name
          when 'capability.zServer'
            handle_z_server(ele, node)
          when 'capability.pRSM'
            handle_capability(ele, node)  # TODO model-specific handler needed
          when 'link.hosting'
            handle_link_hosting(ele, node)
          else
            raise "unknown child element #{ele.name} on #{zsu.name}"
          end
        }

      end

      def handle_lpar_unit(lu)
        # <systemz:unit.lpar conceptual='true' configurationUnit='false' displayName='LPAR'
        #   name='LPARUnit_8549'>

        # TODO Create a Weaver entity (Component?) to hold capability values
        component = entity_from_unit(lu, Lpar)

        lu.elements.each { |ele|
          case ele.name
          when 'capability.lpar'
            handle_lpar(ele, component)
          else
            handle_unit_subtag(ele, lu, component)
          end
        }
      end

      def handle_z_cp_unit(zcpu)
        # <systemz:unit.zCP conceptual='true' configurationUnit='true' displayName='ZIIP' name='ZCPUnit_8549'>

        node = entity_from_unit(zcpu, Node)

        zcpu.elements.each { |ele|
          case ele.name
          when 'capability.zCP'
            handle_z_cp(ele, node)
          when 'requirement'
            handle_requirement(ele, node)
          else
            raise "unknown child element #{ele.name} on #{zcpu.name}"
          end
        }
      end

      def handle_power_server_unit(psu)
        # <server:unit.powerServerUnit displayName='POWER Server' name='PowerServerUnit_996340'>

        # TODO Create a Weaver entity (Node?) to hold capability values
        node = entity_from_unit(psu, Server)

        psu.elements.each { |ele|
          case ele.name
          when 'capability.powerServer'
            handle_power_server(ele, node)
          when 'link.hosting'
            handle_link_hosting(ele, node)
          else
            raise "unknown child element #{ele.name} on #{psu.name}"
          end
        }

      end

      def handle_z_operating_system_unit(zosu)
        os = entity_from_unit(zosu, Node)

        zosu.elements.each { |ele|
          case ele.name
          when 'capability.zOS'
            # Capability name is inherited if not present on Unit
            os.name = ele.attributes['displayName'] unless !os.name.nil?
            handle_z_operating_system(ele, os)
          when 'requirement'
            handle_requirement(ele, os)
          when 'link.hosting'
            handle_link_hosting(ele, os)
          else
            raise "unknown child element #{ele.name} on #{zosu.name}"
          end
        }
      end

      def handle_aix_operating_system_unit(aosu)
        os = entity_from_unit(aosu, Node)

        aosu.elements.each { |ele|
          case ele.name
          when 'capability.aixOperatingSystem'
            # Capability name is inherited if not present on Unit
            os.name = ele.attributes['displayName'] unless !os.name.nil?
            handle_aix_operating_system(ele, os)
          when 'capability.unixDirectory'
            handle_linux_directory(ele, os)
          else
            handle_unit_subtag(ele, aosu, os)
          end
        }
      end

      def handle_generic_operating_system_unit(osu)
        os = entity_from_unit(osu, Node)

        osu.elements.each { |ele|
          case ele.name
          when 'capability.operatingSystem'
            handle_generic_operating_system(ele, os)
          when 'capability.unixDirectory'
            handle_linux_directory(ele, os)
          else
            handle_unit_subtag(ele, osu, os)
          end
        }
      end

      def handle_linux_operating_system_unit(losu)
        os = entity_from_unit(losu, Node)

        losu.elements.each { |ele|
          case ele.name
          when 'capability.suseLinuxOperatingSystem'
            # Capability name is inherited if not present on Unit
            os.name = ele.attributes['displayName'] unless !os.name.nil?
            handle_suse_linux_operating_system(ele, os)
          when 'capability.redhatLinuxOperatingSystem'
            handle_redhat_linux_operating_system(ele, os)
          when 'capability.unixDirectory'
            handle_linux_directory(ele, os)
          when 'capability.linuxOperatingSystem'
            # Capability name is inherited if not present on Unit
            os.name = ele.attributes['displayName'] unless !os.name.nil?
            handle_linux_operating_system(ele, os)
          else
            handle_unit_subtag(ele, losu, os)
          end
        }
      end

      def handle_contract_explicit(ce)
        # Do nothing, we don't currently support unit and attribute visibility and augmentation

        ce.elements.each { |ele|
          case ele.name
          when 'exported.unit'
            # Ignore exports
          else
            raise "unknown element #{ele.name} on #{ce.name}"
          end
        }
      end

      def handle_constraint_networkcommunication(ele, l)
        # <constraint:constraint.networkcommunication name='...' respectLinkDirection='true'/>
        @@logger.debug "TODO network communication from #{l.keys[0]} to #{l.values[0]}"
      end
      
      def handle_constraint_applicationcommunication(ele, l)
        # <constraint:constraint.applicationcommunication name='c_6494'/>
        @@logger.debug "TODO application communication from #{l.keys[0]} to #{l.values[0]}"
      end
      
      def handle_constraint_collocation(ele, l)
        # <constraint:constraint.collocation capabilityType='core:Capability' type='anticollocation'/>
        @@logger.debug "TODO collocation type=#{ele.attributes['type']} from #{l.keys[0]} to #{l.values[0]}"
      end
      
      def handle_attribute_propagation(ap, weaver_object)
        # <constraint:constraint.attributePropagation displayName='Propagate Version Information'
        #  name='c0' sourceAttributeName='versionString' sourceObjectURI='cap1'
        #  targetAttributeName='wasVersion' targetObjectURI='WebSphereSystem'/>

        # TODO.  We may not be ready to resolve sourceObjectURI until later so this method should
        # merely save the info, then later when we are doing hosting and member links we should link
        # up attributes too.

        @@logger.debug "TODO attribute propagation from #{ap.attributes['sourceAttributeName']} to #{ap.attributes['targetAttributeName']}"

      end

      def handle_constraint_other(constraint, weaver_object)
        @@logger.debug "Ignoring constraint #{constraint.name}"
      end

      def handle_constraint_range(cr, weaver_object)
        # <constraint:constraint.range attributeName='osVersion' maxValue='' minValue='' name='osVersionVersio'/>

        @@logger.debug "TODO constraint.range on #{cr.attributes['attributeName']} [#{cr.attributes['minValue']}..#{cr.attributes['maxValue']}]"
      end

      def handle_constraint_version(cv, weaver_object)
        # <constraint:constraint.version attributeName='versionString' name='versionStringVe' uuid='_3juRSE6IEeG0Fozkp1X9yQ' value='6+'/>

        @@logger.debug "TODO constraint.version on #{cv.attributes['attributeName']} [#{cv.attributes['value']}"
      end

      def handle_constraint_deferred_hosting(dh, parent_constraint, weaver_object)
        # TODO support
        # <constraint:constraint.deferredHosting name='c_897996'/>
        # instead of doing nothing

        handle_child_constraints(dh.elements, dh, weaver_object)
      end

      def handle_constraint_capacity(cc, parent_constraint, weaver_object)
        # TODO support
        # <constraint:constraint.capacity displayName='Capacity Constraint' dmoType='server:Server' name='c_2320'>
        # instead of doing nothing

        handle_child_constraints(cc.elements, cc, weaver_object)
      end

      def handle_capacity_restriction_constraint(crc, parent_constraint, weaver_object)
        # TODO handle
        # <constraint:constraint.capacityRestriction attributeName='memorySize' exclusive='true' name='c_8231' value='30MB'/>

        handle_child_constraints(crc.elements, crc, weaver_object)
      end

      def handle_child_constraints(elements, constraint, weaver_object)
        elements.each { |ele|
          case ele.name
          when 'constraint.capacityRestriction'
            handle_capacity_restriction_constraint(ele, constraint, weaver_object)
          else
            raise "unknown child_constraint #{ele.name} on #{constraint.name}; #{ele}"
          end
        }
      end

      def handle_stereotype(stereotype, weaver_entity)
        # <core:stereotype keyword='StorageSubsystem' profile='newProfile'/>

        # Ignore stereotypes
      end

      def handle_runLevel(runLevel, weaver_entity)
        # <appliance:runLevel>3</appliance:runLevel>

        # TODO Do something
      end

      def handle_specificationLevel(specificationLevel, weaver_entity)
        # <specificationLevel>CLASSIC_IBM</specificationLevel>

        # TODO Do something
      end

      def handle_capability(capability, weaver_entity, processed_attributes = {})
        raise "weaver entity problem" unless weaver_entity.kind_of?(Base)

        # Convert Zephyr attributes into Weaver attributes
        capability.attributes.each { |key, val|
          if !core_attribute_name(key) && !processed_attributes.member?(key)
            attr = Attribute.new(key.symbolize, weaver_entity)
            attr.default = val
            attr.type = :string
            attr.required = false
            attr.value = val
            weaver_entity.amd[key.symbolize] = attr
          end
        }

        # Handle Zephyr child elements (metadata and constraints)
        capability.elements.each { |ele|
          case ele.name
          when 'attributeMetaData'
            handle_amd(ele, weaver_entity)
          when 'extendedAttribute'
            handle_extended_attribute(ele, weaver_entity)
          when 'constraint.range'
            handle_constraint_range(ele, weaver_entity)
          when 'dependsOnService'
            @@logger.debug "Ignoring <#{ele.expanded_name}>"
          when 'annotation'
            handle_annotation(ele, weaver_entity)
          when 'constraint.version'
            handle_constraint_version(ele, weaver_entity)
          when /^constraint\./
            handle_constraint_other(ele, weaver_entity)
          when 'specificationLevel'
            handle_specificationLevel(ele, weaver_entity)
          when 'runLevel'
            handle_runLevel(ele, weaver_entity)
          when 'stereotype'
            handle_stereotype(ele, weaver_entity)
          when 'ovfProperties'
            @@logger.debug "Ignoring OVF annotation; attribs=#{ele.attributes}"
          else
            raise "unknown element #{ele.name} on Weaver Entity #{weaver_entity}"
          end
        }
      end
      
      def handle_port_forward(capability, og)
        # <net:capability.portForward endingPort='-1' forwardTCP='true' forwardUDP='false' ipV4Address='0.0.0.0/0' name='portForward' startingPort='-1'/>
        
        if capability.attributes['forwardTCP'] == 'true'
          startingPort = capability.attributes['startingPort']
          endingPort = capability.attributes['endingPort']
          ipV4Address = capability.attributes['ipV4Address']
          if !ipV4Address.nil? && ipV4Address != '0.0.0.0/0'
            @@logger.warn "Ignoring UDP firewall netmask #{ipV4Address}"
          end
          
          if startingPort == endingPort
            og.ports.push << startingPort.to_i unless startingPort.to_i == -1 
          else
            port_range = startingPort.to_i .. endingPort.to_i
            og.ports.push << port_range
          end
        end
        
        if capability.attributes['forwardUDP'] == 'true'
          @@logger.warn "Ignoring UDP firewall rule"
        end
      end

      def handle_z_cp(capability, weaver_entity)
        # <systemz:capability.zCP cpType='ZIIP' linkType='any' name='zCP'/>

        # TODO Save cpType

        handle_capability(capability, weaver_entity)
      end

      def handle_was_system(capability, component)
        # <was:capability.wasSystem installedLocation='C:\' linkType='any'
        #    name='WebSphereSystem' productName='IBM WebSphere(R) Application Server'
        #    wasHome='C:\Program Files\IBM\WebSphere\AppServer' wasVersion='6.1'/>

        component.version = capability.attributes['wasVersion']
        # TODO Save wasHome, installedLocation
      end

      def handle_db2_system(capability, component)
        # <db2Deploy:capability.db2System db2Version='8.2' displayName='DB2 System' linkType='any' name='DB2 System'/>

        component.version = capability.attributes['db2Version']
      end

      def handle_db2_jdbc_driver(capability, component)
        # <db2Deploy:capability.db2JdbcDriver classname='COM.ibm.db2.jdbc.DB2ConnectionPoolDataSource'
        #   classpath='/IBM/SQLLIB/db2java.zip' displayName='DB2 Legacy CLI-based Type 2 JDBC Driver'
        #   jdbcType='2' linkType='dependency' name='DB2 Legacy CLI-based Type 2 JDBC Driver'/>

        # TODO Record custom DB2 JDBC driver attributes

        handle_capability(capability, component)
      end

      def handle_software_install(capability, component)
        raise "NilArgumentException" unless !component.nil?

        # <genericsoftware:capability.softwareInstall installedLocation='/opt/IBM/ITM' linkType='any'
        #    name='SoftwareInstall' productName='IBM Tivoli Monitoring Agent' publisher='IBM Tivoli'/>

        component.name = capability.attributes["productName"] unless capability.attributes["productName"].nil? 
        # TODO We have nowhere to store 'publisher'
        # TODO We have nowhere to store 'installedLocation'

        handle_capability(capability, component, Set.new([ 'productName' ]))
      end

      def handle_version(version_capability, component)
        # <genericsoftware:capability.version name="version" linkType="any" versionString="1.0"/>

        component.version = version_capability.attributes['versionString']

        handle_capability(version_capability, component)
      end

      def handle_redhat_user(user_capability, user)
        # <os:capability.redhatLinuxLocalUser displayName='root' linkType='dependency'
        #   name='RedhatLinuxLocalUser' userId='root' userPassword='hK/A0uZGH02s'>
        handle_local_user(user_capability, user)  # TODO add support for optional Redhat-only params if we can find an example in the wild
      end

      def handle_local_user(user_capability, user)
        # TODO support
        # <os:capability.suseLinuxLocalUser HomeDir='/clouduser' linkType='dependency'
        #   name='SUSELinuxLocalUser' userId='idcuser' userPassword='eZzfNatir+xCDy5n4PUbxlQ='>

        # TODO How do I set HomeDir?
        user.amd[:username].value = user_capability.attributes['userId']
        user.amd[:password].value = :unknown # We don't support decrypting Zephyr passwords

        handle_capability(user_capability, user, Set.new([ 'name', 'password' ]))
      end

      def handle_run_operation(run_op, op)
        #   <appliance:capability. name="operation" operationName="New Step"
        #      generated="false" parameterStyle="environment" runAt="imageBuild"
        #      interpreter="/bin/sh" runAsUser="root" runCommands="dummy.sh" workingDirectory="/tmp"/>

        @@logger.debug "Processing RunOperation #{run_op.attributes['runCommands']}"
        op.run_at = run_op.attributes['runAt']
        op.parameter_style = run_op.attributes['parameterStyle']
        op.interpreter = run_op.attributes['interpreter']
        op.run_as = run_op.attributes['runAsUser']
        op.run_commands = run_op.attributes['runCommands']
        op.working_directory = run_op.attributes['workingDirectory']

        handle_capability(run_op, op)
      end

      def handle_x86_server(x86_server, server)
        # <server:capability.x86Server cpu.architecture='intel' cpu.count='2'
        #   cpuArchitectureWidth='64-bit' cpuCoresInstalled='2' cpuManufacturer='AMD'
        #   displayName='x86 Server' linkType='any' memory.size='2000000' name='x86 Server'>

        # TODO Preserve this information?

        handle_capability(x86_server, server)
      end

      def handle_server(zephyr_server, server)
        # <server:capability.server linkType='any' name='server'/>

        # TODO Preserve this information?

        handle_capability(zephyr_server, server)
      end

      def handle_z_server(z_server, server)
        # <systemz:capability.zServer cpu.architecture='systemz' displayName='z Server' linkType='any' name='z Server'>

        # TODO Preserve this information?

        handle_capability(z_server, server)
      end

      def handle_power_server(x86_server, server)
        # <server:capability.x86Server cpu.architecture='intel' cpu.count='2'
        #   cpuArchitectureWidth='64-bit' cpuCoresInstalled='2' cpuManufacturer='AMD'
        #   displayName='x86 Server' linkType='any' memory.size='2000000' name='x86 Server'>

        # TODO Preserve this information?

        handle_capability(x86_server, server)
      end

      def handle_z_operating_system(z_capability, os)
        # <systemz:capability.zOS linkType='any' name='zOS'/>

        os.type = :zOS

        handle_capability(z_capability, os)
      end

      def handle_lpar(capability, weaver_entity)
        # <systemz:capability.lpar cpu.architecture='systemz' linkType='any' name='LPAR'/>

        handle_capability(capability, weaver_entity)
      end

      def handle_aix_operating_system(aix_capability, node)
        # <os:capability.aixOperatingSystem displayName='AIX' linkType='any' name='AIX' os.type='AIX' os.version='5L'/>

        os = node.os
        os.type = :linux
        os.flavor = :aix
        os.version = aix_capability.attributes['os.version']
        if os.version.nil?
          os.version = :unknown
        end

        handle_capability(aix_capability, node, Set.new([ 'os.version', 'os.type' ]))
      end

      def handle_generic_operating_system(capability, os)
        handle_linux_operating_system(capability, os)
      end

      def handle_linux_operating_system(linux_capability, node)
        # TODO support
        # <os:capability.suseLinuxOperatingSystem Desktop='GNOME' displayName='SUSE Linux Server'
        # kernelArchitecture='x86' kernelWidth='32-bit' linkType='any' manufacturer='Xen'
        # name='SUSELinux' os.distribution='SuSE' os.servicepack='0' os.type='Linux' os.version='11'/>
        
        os = node.os

        os_dist = linux_capability.attributes['os.distribution']
        if os_dist.nil? || os_dist.size == 0
          os.type = :linux
        else
          os_info = @os_handler.match_os(os_dist)
          #          raise "No OS match for #{linux_capability}" unless !os_info.nil?
          if !os_info.nil?
            os.type = os_info[:type]
            os.flavor = os_info[:flavor]
          else
            os.type = :linux
          end
        end
        os.version = linux_capability.attributes['os.version']
        if os.version.nil?
          os.version = :unknown
        end

        os.architecture = @os_handler.match_architecture(linux_capability.attributes['kernelArchitecture'])

        handle_capability(linux_capability, node, Set.new([ 'os.type', 'os.distribution', 'os.version', 'kernelArchitecture' ]))
      end

      def handle_linux_directory(linux_capability, os)
        # <os:capability.unixDirectory FixedPath='/tmp' displayName='directory' linkType='any' name='directory'>
        #        raise "Unsupported capability #{linux_capability}"

        handle_capability(linux_capability, os)
      end

      def handle_redhat_linux_operating_system(linux_capability, os)
        handle_linux_operating_system(linux_capability, os)
      end

      def handle_suse_linux_operating_system(linux_capability, os)
        handle_linux_operating_system(linux_capability, os)
      end

      def handle_extended_attribute(extendedAttribute, weaver_entity)
        # <core:extendedAttribute name='sharedResourcesDirectory'>
        #   <core:data xsi:type='xsd:string'>/opt/IBM/IBMIMShared</core:data>
        # </core:extendedAttribute>
        name = extendedAttribute.attributes['name']
        extendedAttribute.elements.each do |ele|
          if ele.name == "data"
            attr = Attribute.new(name.symbolize, weaver_entity)
            val = ele.text
            attr.value = val
            weaver_entity.amd[attr.id] = attr
            return
          end
        end
        raise "Unimplemented: No <core:data> under <extendedAttribute> '#{name}'"
      end

      def handle_amd(amd, weaver_entity)
        # TODO commented out - don't see value in displaying this information yet.
        id = amd.attributes['attributeName'].symbolize
        encrypted = amd.attributes['encrypted'] == 'true'
        label = amd.attributes['label']
        mutable = amd.attributes['mutable']

        attr = weaver_entity.amd[id]
        if attr.nil?
          attr = Attribute.new(id, weaver_entity)
          weaver_entity.amd[id] = attr
        end
        attr.name = label
        attr.mutable = mutable
        attr.password = encrypted
      end

      def handle_reference(reference, weaver_entity)
        # <core:reference displayName='TDS DB' dmoType='database:DatabaseDefinition' linkType='dependency' name='r1' use='required'>
        #     <core:link.dependency name='r1ToDatabaseDefinition_2091' source='/SoftwareComponent_2468/r1' target='/DB2DatabaseUnit_1275/DatabaseDefinition_2091'/>
        #     <database:interface.ddl schemaUri='TDS DB'/>
        #   </core:reference>

        # TODO Handle?
        @@logger.debug "Ignoring <reference>"
      end

      def handle_annotation(annotation, weaver_entity)
        # <core:annotation context='abstraction_bundle'>
        #     <core:detail key='isAbstraction' value='true'/>
        #   </core:annotation>

        if annotation.attributes['context'] == 'parameter_annotation'
          handle_parameter_annotation(annotation, weaver_entity)
        else
          # TODO Model annotations in Weaver through some kind of extension mechanism (e.g. an "annotations" Hash on Base)
          @@logger.debug "Ignoring <annotation> context '#{annotation.attributes['context']}' on #{weaver_entity.class}"
        end
      end

      def handle_parameter_annotation(annotation, weaver_entity)
        # <core:detail key="parameter" value="config_msg"/>
        # <core:detail key="direction" value="input"/>

        parameter, direction = nil
        annotation.elements.each { |detail_ele|
          keyval = detail_ele.attributes['key']
          if keyval == 'parameter'
            parameter = detail_ele.attributes['value']
          elsif keyval == 'direction'
            direction = detail_ele.attributes['value']
          else
            @@logger.debug "Ignoring parameter annotation #{keyval}=#{detail_ele.attributes['value']}"
          end
        }

        if (direction == 'input' || direction == 'inout') && deploy_operation?(weaver_entity)
          # If this annotation is done upon a <appliance:capability.> there may be a serviceName= attribute
          serviceName = annotation.parent.attributes['serviceName']
          attribute = Attribute.new(parameter.symbolize, weaver_entity)
          attribute.required = true
          attribute.mutable = true
          attribute.label = "TODO" # There is an AMD with the label, but it may not have been reached yet
          attribute.name = serviceName.nil? ? parameter : "#{serviceName}.#{parameter}"
          attribute.group = :deploy_param
          @parameters[attribute.id] = attribute
        else
          @@logger.debug "Ignoring parameter #{parameter} [#{direction}]"
        end

      end

      def handle_service(service, weaver_entity)
        # examples:
        
        # <database:service.databaseDefinition displayName='TDS DB' linkType='any' name='DatabaseDefinition_2091'>
        #     <database:interface.ddl schemaUri='TDS DB'/>
        #   </database:service.databaseDefinition>
        
        # <j2eeDeploy:service.ws displayName="AmazonS3" name="WebService_1719" linkType="any" protocol="http">
        #  <j2eeDeploy:interface.wsdl typeUri="/AmazonEC2/templates/com/amazon/aws/storage/wsdl/AmazonS3.wsdl#AmazonS3"/>
        # </j2eeDeploy:service.ws>
        
        # <core:service displayName='Interface Spec' linkType='any' name='Service_241687'>
        #   <core:attributeMetaData attributeName='Release' mutable='true' parameter='true'/>
        #   <core:extendedAttribute name='Release'>
        #     <core:data xsi:type='xsd:string'>R</core:data>
        #   </core:extendedAttribute>
        #   ...
        #   <core:stereotype keyword='Interface' profile='newProfile'/>
        # </core:service>
        
        # <j2eeDeploy:service.session displayName='Catalog' ejbName='Catalog' jndiName='ejb/Catalog' name='Catalog' protocol='RMI' version='2.0' vicinity='remote'>
        #  <was:constraint.wasJNDIBindingConstraint jndiName='ejb/Catalog' name='WasJNDIBindingConstraint_373'/>
        #  <j2eeDeploy:interface.java interface='edu.nyu.pdsg.tpcw.ejb.catalog.Catalog'/>
        #  <j2eeDeploy:homeInterface interface='edu.nyu.pdsg.tpcw.ejb.catalog.CatalogHome'/>
        # </j2eeDeploy:service.session>
        
        # <j2eeDeploy:service.servlet name='HomeServlet' protocol='http' version='1.0.0'>
        #  <j2eeDeploy:interface.url/>
        #  <j2eeDeploy:urlPattern>/home</j2eeDeploy:urlPattern>
        # </j2eeDeploy:service.servlet>
        
        # <j2eeDeploy:service.entity displayName='IdManager' ejbName='IdManager' jndiName='ejb/IdManager' managed='true' name='IdManager' protocol='RMI' version='2.0' vicinity='local'>
        #  <was:constraint.wasJNDIBindingConstraint jndiName='ejb/IdManager' name='WasJNDIBindingConstraint_6476'/>
        #  <j2eeDeploy:interface.java interface='edu.nyu.pdsg.tpcw.ejb.idmanager.IdManager'/>
        #  <j2eeDeploy:homeInterface interface='edu.nyu.pdsg.tpcw.ejb.idmanager.IdManagerHome'/>
        # </j2eeDeploy:service.entity>

        if service.name == 'service.databaseDefinition' || service.name == 'service.ws' || 
            service.name == 'service' || service.name == 'service.session' || 
            service.name == 'service.servlet' || service.name == 'service.entity'
          @@logger.debug "Ignoring service #{service.name}"
        else
          raise "Unknown Service Type #{service}" ## @@@ TODO REMOVE
        end
      end

      def handle_import(import)
        # <core:import displayName='Logical Operational Model' mutable='true' name='Logical Operational Model' namespace='com.ibm.tellersystem.topology' pattern='Logical Operational Model'>
        #   <core:config.topology name='default'/>
        # </core:import>

        @@logger.debug "Ignoring <import>"
        #        raise "Unsupported import #{import}"
      end

      def handle_requirement(requirement, weaver_entity)
        # TODO support
        # <core:requirement dmoType='os:OperatingSystem' linkType='hosting' name='Host on an Operating System' use='required'/>
        #        raise "Unsupported requirement #{requirement}"
      end

      def handle_link_member(link, weaver_entity)
        # <core:link.member name='LPARUnit_1453ContainsZCPUnit_7646' source='#/LPARUnit_8549' target='#/ZCPUnit_8549'/>

        @member_links.push(parse_link(link, weaver_entity))
      end

      def handle_link_constraint(link, weaver_entity)
        # <core:link.constraint name='cl_897995' source='#/WasSystemUnit_263533'
        #   target='#/X86ServerUnit_992066'>

        l = parse_link(link, weaver_entity)

        # Handle the constraints attached to the link
        link.elements.each { |ele|
          case ele.name
          when 'constraint.deferredHosting'
            handle_constraint_deferred_hosting(ele, nil, l)
          when 'constraint.attributePropagation'
            handle_attribute_propagation(ele, l)
          when 'constraint.networkcommunication'
            handle_constraint_networkcommunication(ele, l)
          when 'constraint.applicationcommunication'
            handle_constraint_applicationcommunication(ele, l)
          when 'constraint.collocation'
            handle_constraint_collocation(ele, l)
          when /^constraint\./
            handle_constraint_other(ele, l)
          else
            raise "unknown element #{ele.name} on #{link.name}"
          end
        }

        @constraint_links.push(l)
      end

      def handle_link_realization(link, weaver_entity = nil)
        # <core:link.realization name='DatabaseSystemUnit_517802RealizedByDB2SystemUnit_545738'
        #   source='#/DatabaseSystemUnit_517802' target='#/DB2SystemUnit_545738'/>

        l = parse_link(link, weaver_entity)
        @realization_links.push(l)

        # TODO Save weaver_entity, if any
      end

      def handle_link_dependency(link, weaver_entity)
        # <core:link.dependency name='r2ToService_241687' source='blue.harmony.architecture.GT Apps:/SoftwareComponent_30171/r2' target='/SoftwareComponent_80734/Service_241687'/>
        @@logger.debug "Ignoring dependency link; source='#{link.attributes['source']}' target='#{link.attributes['target']}'"
      end
      
      def handle_link_hosting(link, weaver_entity)

        # Use an Array of Hash because we sometimes have {M:N} relationships (not {1:N})
        @hosting_links.push(parse_link(link, weaver_entity))
      end
      
      def handle_file_artifact(image_file_ele, weaver_entity)
        # <appliance:artifact.imageFile deployLocation='/etc/cloud/setupvnc.sh' executable='true' name='imageArtifact'>
        #  <core:fileURI>artifacts/setupvnc.sh</core:fileURI>
        # </appliance:artifact.imageFile>
        
        if weaver_entity.kind_of?(RunOperation)
          deployLocation = image_file_ele.attributes['deployLocation'] 
          executable_flag = image_file_ele.attributes['executable'] == 'true'
          uri = image_file_ele.elements[1].get_text()
          weaver_entity.artifacts.push({ :deployLocation=>deployLocation, :executable_flag=>executable_flag, :uri=>uri})
        else
          @@logger.debug "TODO <artifact.imageFile> on a #{weaver_entity.class}"
        end
      end

      # Returns a Hash whose only entry is source-symbol pointing to target-symbol
      def parse_link(link, weaver_entity)
        source_id = link.attributes['source']
        if source_id.nil?
          source_sym = weaver_entity.id
        else
          # In Ruby 1.8.x, String[fixnum] returns a Fixnum of the ASCII code, in 1.9.x a String of the character
          # String[fixnum,fixnum] returns a String in both versions.
          if source_id[0,1] == '#'
            source_id = source_id.slice(1 .. -1)
          end
          if source_id[0,1] == '/'
            source_id = source_id.slice(1 .. -1)
          end
          source_sym = source_id.symbolize
        end

        # I found a few topologies with no target and I tolerate that here
        target_id = link.attributes['target'] || 'error_missing_target'
        # In Ruby 1.8.x, String[fixnum] returns a Fixnum of the ASCII code, in 1.9.x a String of the character
        # String[fixnum,fixnum] returns a String in both versions.
        if target_id[0,1] == '#'
          target_id = target_id.slice(1 .. -1)
        end
        if target_id[0,1] == '/'
          target_id = target_id.slice(1 .. -1)
        end

        return { source_sym => target_id.symbolize }
      end

      def remember_entity(zephyr_unit, weaver_entity)
        id = zephyr_unit.attributes['name'].symbolize
        if @entities.member?(id)
          @@logger.warn "Duplicate ID #{id} on #{zephyr_unit.name}; previously used for a #{@entities[id].class}"
          return
        end

        @entities[id] = weaver_entity

        if conceptual_unit(zephyr_unit)
          @conceptuals[id] = weaver_entity
        end

        if installed(zephyr_unit)
          @installed_units.add(id)
        elsif to_be_installed(zephyr_unit)
          @to_be_installed_units.add(id)
        end
      end

      def installed(element)
        initial_state = element.attributes['initInstallState']
        return initial_state == "installed" || (initial_state.nil? && !conceptual_unit(element)) || element.name == "unit.virtualSystem"
      end

      def to_be_installed(element)
        initial_state = element.attributes['initInstallState']
        goal_state = element.attributes['goalInstallState']
        return initial_state == "uninstalled" && goal_state == "installed"
      end

      def conceptual_unit(element)
        is_conceptual = element.attributes['conceptual']
        return is_conceptual == "true"
      end

      def link_entities
        @realization_links.each { |hash|
          realize_ids(hash.keys[0], hash.values[0])
        }

        @hosting_links.each { |hash|
          host_ids(hash.keys[0], hash.values[0])
        }

        @member_links.each { |hash|
          group_ids(hash.keys[0], hash.values[0])
        }

        @constraint_links.each { |hash|
          @@logger.debug "Unhandled constraint link #{hash}"
        }

        @parameters.each { |id, parameter|
          expose_parameter(parameter)
        }
      end

      def expose_parameter(parameter)
        # Pick up the label and default value from the scope
        id = parameter.id
        parameter.label = parameter.scope.amd[id].name
        parameter.default = parameter.scope.amd[id].value
        parameter.type = :string  # TODO type from topology <core:data xsi:type="xsd:string" ...>

        # TODO Replace the original parameter.scope.unresolved_attributes with a reference to the cloud deploy parameter?
      end

      def realize_ids(source_id, target_id)
        raise "source_id must be Symbol" unless source_id.kind_of?(Symbol)
        raise "target_id must be Symbol" unless  target_id.kind_of?(Symbol)

        if @entities[source_id].nil?
          #          raise "Unknown realization source id #{source_id}, ids are #{@entities.keys}"
          # We only dump the error, rather than failing, because we don't process <import> and
          # the target could be an imported node.  An example of this is
          # 'com.ibm.tellersystem.topology.Logical Operational Model:/NodeUnit_8206'
          @@logger.info "Unknown source id #{source_id}, ids are #{@entities.keys}"
          return
        end

        if @entities[target_id].nil?
          #          raise "Unknown realization target id #{target_id}, ids are #{@entities.keys}"
          # We only dump the error, rather than failing, because we don't process <import> and
          # the target could be an imported node.  An example of this is
          # 'com.ibm.tellersystem.topology.Logical Operational Model:/NodeUnit_8206'
          @@logger.info "Unknown target id #{target_id}, ids are #{@entities.keys}"
          return
        end

        realize(@entities[source_id], @entities[target_id])
      end

      def group_ids(source_id, target_id)
        if @entities[source_id].nil?
          # We only dump the error, rather than failing, because we don't process <import> and
          # the source could be an imported node.  An example of this is
          # 'amgro_analysis_logicalmodel:/location4'
          @@logger.info "Unknown group source id #{source_id}, ids are #{@entities.keys}"
        end
        if @entities[target_id].nil?
          #          raise "Unknown target id #{target_id}, ids are #{@entities.keys}"
          # We only dump the error, rather than failing, because we don't process <import> and
          # the target could be an imported node.  An example of this is
          # 'IBMCoCBaseImage.OSNode.OSNode_semantic:/X86ServerUnit_344564'
          @@logger.info "Unknown group target id #{target_id}, ids are #{@entities.keys}"
          return
        end

        group(@entities[source_id], @entities[target_id])
      end

      def host_ids(source_id, target_id)
        if @entities[source_id].nil?
          # We only dump the error, rather than failing, because we don't process <import> and
          # the source could be an imported node.  An example of this is
          # 'application_profile:/wasclusterunit_184500'
          @@logger.info "Unknown host source id #{source_id}, ids are #{@entities.keys.to_a}"
          return
        end
        if @entities[target_id].nil?
          # We only dump the error, rather than failing, because we don't process <import> and
          # the target could be an imported node.  An example of this is
          # 'blue_harmony_architecture_gt_apps:/softwarecomponent_964687'
          @@logger.info "Unknown host target id #{target_id}, ids are #{@entities.keys}"
          return
        end

        host(@entities[source_id], @entities[target_id], @conceptuals.key?(target_id))
      end

      def realize(source, target)
#        @@logger.debug "realize source=#{source.id} (#{source.object_id}), target=#{target.id} (#{target.object_id})"
        if source.kind_of?(IgnoredUnit)
          @@logger.debug "Unhandled realization link source #{source.id}"
          return
        end

        if source.kind_of?(Component) || source.kind_of?(Server) || source.kind_of?(Group)
          @realizations[source] = target
#          @@logger.debug "Zephyr realization of #{source.class} ignored because Weaver links only Nodes (#{source.id} to #{target.id})"
          return
        end

        source.realized_by << target
      end

      def host(source, target, is_conceptual)
#        @@logger.debug "host source=#{source.id} (#{source.class}), target=#{target.id}"
        target = most_realized(target)
        
        if source.kind_of?(Node) && target.kind_of?(User)
          # do nothing, we will expose users a different way # TODO link users to infrastructure and specific nodes
        elsif source.kind_of?(Node) && target.kind_of?(Component)
          target.scope = source
          if is_conceptual
            # Turn conceptual Component into a Requirement
            reqs = { :product => target.name, :version => target.version }
            prod_req = Weaver::Dsl::Requires::ProductRequirement.new(source, reqs)
            source.requirements << prod_req
          else
            # Don't host Activation Engine on a node
            source.components[target.id] = target unless target.name =~ /Activation Engine/ || target.name =~ /ConfigIcon/
          end
        elsif source.kind_of?(Server)
          # Ignore, we are not saving Zephyr 'server' information at this time
        elsif source.kind_of?(Lpar)
          # Ignore, we are not saving Zephyr 'LPAR' information at this time
        elsif source.kind_of?(IgnoredUnit) || target.kind_of?(IgnoredUnit)
          # Ignore, we are not saving Zephyr this information
        elsif source.kind_of?(Component) && target.kind_of?(Component)
          @@logger.debug "Trying to host component #{source.id} on #{target.id}"
          target.scope = source
          if @conceptuals.key?(target.id)
            # TODO Create a requirement when we see a conceptual target
            # source.requirements[target.id] = target
          else
            source.components[target.id] = target
          end
        elsif source.kind_of?(Component) && target.kind_of?(Group)
          # Ignore, currently we generate Groups on the Infrastructure rather than within the Component that defined them in Zephyr
        elsif source.kind_of?(Component) && target.kind_of?(User)
          @@logger.debug "Ignoring User #{source.id} hosted on Component #{target.id}"
        else
          raise "Unimplemented hosting link from #{source.class} :#{source.id} to #{target.class} :#{target.id}"
        end
      end

      def group(source, target)
        if source.kind_of?(Lpar)
          # Ignore, we are not saving Zephyr 'LPAR' information at this time
        elsif source.kind_of?(IgnoredUnit) || source.kind_of?(UnmodeledConcept)
          # Ignore, we are not saving Zephyr other information at this time
        elsif target.kind_of?(IgnoredUnit)
          # Ignore, we are not saving Zephyr other information at this time
        elsif source.kind_of?(OperationGroup) && target.kind_of?(RunOperation)
          source.scripts.push(target)
        elsif target.kind_of?(OperationGroup) || source.kind_of?(OperationGroup)
          # Ignore, we are not saving Zephyr Operation Group independently of the Virtual System
        else
          @@logger.warn "Unimplemented member/grouping link from #{source.class} to #{target.class}"
        end
      end

      def entity_from_unit(zephyr_unit, clazz)
        if clazz.kind_of?(Class)
          val = clazz.new(zephyr_unit.attributes['name'].symbolize, nil)
        else
          val = @root_entity.prototype_instance(clazz, zephyr_unit.attributes['name'].symbolize, nil, nil)
        end
        val.name = zephyr_unit.attributes['displayName']
        val.description = zephyr_unit.attributes['description']
        remember_entity(zephyr_unit, val)
        return val
      end
      
      def most_realized(entity)
        next_entity = @realizations[entity]
        while !next_entity.nil? do
          entity = next_entity
          next_entity = @realizations[entity]
        end
        return entity
      end

      def core_attribute_name(attr)
        if Zephyr::CORE_ATTRIBUTES.member?(attr)
          #          @@logger.debug "'#{attr}' is a core attribute"
          return true
        end

        #        @@logger.debug "'#{attr}' is NOT a core attribute"
        return false
      end
      # :startdoc:

      private

      def deploy_operation?(weaver_entity)
        return weaver_entity.kind_of?(RunOperation) && weaver_entity.run_at == 'imageDeploy'
      end

    end

    # @private
    # A dummy class to hold Servers, currently we throw away their information
    class Server < Weaver::Core::Base # :nodoc: all
    end

    # @private
    # A dummy class to hold Lpar, currently we throw away their information
    class Lpar < Weaver::Core::Base # :nodoc: all
    end

    # @private
    # A dummy class to hold OperationGroup, we will hold it on behalf of the Virtual System
    class OperationGroup < Weaver::Core::Base # :nodoc: all
      attr_accessor :ports
      attr_accessor :scripts
      
      def initialize(id, parent)
        super
        @ports = []
        @scripts = []
      end
    end

    # @private
    # A dummy class to hold Other types, currently we throw away their information except description
    class IgnoredUnit < Weaver::Core::Base # :nodoc: all
    end
    
    # @private
    # A dummy class that generates a concept for a concept unmodeled in Weaver
    class UnmodeledConcept < Weaver::Core::Base # :nodoc: all
      attr_accessor :note
    end

    # @private
    # This class will be transformed into a Weaver Script
    class RunOperation < Weaver::Core::Base # :nodoc: all
      attr_accessor :run_at, :parameter_style, :interpreter, :run_as, :run_commands, :working_directory
      attr_reader :artifacts
      def initialize(id, parent)
        super
        @artifacts = []
      end
    end

  end
end
